# -*- coding: utf-8 -*-
"""
Created on Tue Jun 13 22:38:05 2017

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import os
import sys
import ast
import codecs
import importlib
import numpy as np
import warnings

from re import split as resplit, findall as refindall, sub as resub
from copy import deepcopy

with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    import matplotlib.pyplot as plt
    import matplotlib as mpl

from grapa.graph import Graph
from grapa.curve import Curve
from grapa.mathModule import is_number, stringToVariable
from grapa.curve_inset import Curve_Inset
from grapa.curve_subplot import Curve_Subplot
from grapa.curve_image import Curve_Image # needed for curve casting



FILEIO_GRAPHTYPE_DATABASE = 'Database'
FILEIO_GRAPHTYPE_GRAPH = 'Graph'
FILEIO_GRAPHTYPE_UNDETERMINED = 'Undetermined data type'


class GraphIO(Graph):
    """
    Subclass of Graph handling the file reading and writing.
    Want a separate file for a bad reason (Graph was too big). Thus the class
    is written assuming the self is not a Graph instance, but not a GraphIO.
    By default, 3 different reading modes are possible:
    - directly from data (filename=[[0,1,2,3],[5,2,7,4]])
    - column-organized datafile, with data starting at first row with a number
      in the first field
    - database, with no number in the first column
    File opening strategy:
        1/ readDataFile
            - do some stuff with complementary input
            - read first 3 lines of the file (readDataFileLine123)
            - identify child classes which (_listGraphChildClasses)
            - check if any of Graph child class can handle the file
            - if not, calls readDataFromFileTxt
        2/ readDataFromFileTxt
            - look if file has a number in the first column (_funcReadDataFile)
              if so, call readDataFromFileGeneric
              else call readDataFromFileDatabase
    """
        
    childClasses = None
    # list of child classes of Graph



    def dataFromVariable(self, data, attributes):
        """ Initializes content of object using data given in constructor. """
        self.data.append(Curve(data, {}))
        self.update(attributes, ifAll=False)
        # nevertheless want to import all Curve types, if not already done
        # so Curve will know all its possible subtypes and can tell to GUI
        GraphIO._listGraphChildClasses()


    @classmethod
    def _listGraphChildClasses(cls):
        """
        Returns the child classes of Graph which have implemented file
        reading. Seeps through files with names starting with 'curve' or
        'graph', check the presence of required methods,
        and if so returns the Graph child class.
        """
        # do not reimport everything each time, only first time
        if cls.childClasses is not None:
            return cls.childClasses
        folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'datatypes')
        required = ['FILEIO_GRAPHTYPE', 'isFileReadable', 'readDataFromFile']
        subclasses = []
        for filestart in ['graph', 'curve']:
            for file in os.listdir(folder):
                fileName , fileExt = os.path.splitext(file)
                if (fileExt == '.py' and fileName[:len(filestart)] == filestart
                        and len(fileName) > len(filestart)):
                    end = fileName[len(filestart):]
                    module = importlib.import_module('grapa.datatypes.'+fileName)
                    if not hasattr(module, 'Graph'+end):
                        continue
                    typeM = getattr(module, 'Graph'+end)
                    isValid = True
                    for attr in required:
                        if not hasattr(typeM, attr):
                            isValid = False
                            break
                    if isValid:
                        subclasses.append(typeM)
        Graph.childClasses = subclasses # class constant
        return subclasses

    @classmethod
    def readDataFileLine123(cls, filename):
        """
        Returns the first 3 lines of a file.
        """
        line1, line2, line3 = '', '', ''
        try:
            f = open(filename, 'r')
        except Exception as e: # if file does not exist
            # do NOT want to append Curve object if file empty
            return line1, line2, line3
        # if file exist, try to open first line
        try:
            line1 = f.readline().rstrip(' \r\n\t').replace('ï»¿','')
            try:
                line2 = f.readline().rstrip(' \r\n\t').replace('ï»¿','')
                try:
                    line3 = f.readline().rstrip(' \r\n\t').replace('ï»¿','')
                except Exception:
                    pass
            except Exception:
                pass
        except Exception:
            pass
        f.close()
        return line1, line2, line3

        
    def readDataFile(self, complement=''):
        """
        Reads a file.
        Complement: ...
        Opens file self.filename
        """
        self.isFileContent = True if (isinstance(complement, dict) and
                                      'isfilecontent' in complement and
                                      complement['isfilecontent']) else False
        # attempt reading the file to identify types
        if not self.isFileContent:
            fileName, fileExt = os.path.splitext(self.filename)
            fileExt = fileExt.lower()
        else:
            fileName, fileExt = '', '.txt'
        fileNameBase = fileName.split('/')[-1].split('\\')[-1]
        attributes = {'filename': self.filename}
        if complement != '':
            if isinstance(complement, dict):
                for key in complement:
                    attributes.update({key.lower(): complement[key]})
            else:
                attributes.update({'complement': complement})
        # default attributes of curve
        if 'label' not in attributes:
            attributes.update({'label': fileName.split('/')[-1].split('\\')[-1].replace('_', ' ').replace('  ',' ')})
        # default attributes of Graph
        if 'subplots_adjust' in attributes:
            self.update({'subplots_adjust': attributes['subplots_adjust']})
            del attributes['subplots_adjust']
        else:
            self.update({'subplots_adjust': Graph.DEFAULT['subplots_adjust']})
        if 'fontsize' in attributes:
            self.update({'fontsize': attributes['fontsize']})
            del attributes['fontsize']
        else:
            self.update({'fontsize': Graph.DEFAULT['fontsize']})
        # guess keys of complement which are for Graph
        toDel = []
        for key in complement:
            try:
                if key.lower() in Graph.graphInfoKeys:
                    self.update({key: complement[key]})
                    toDel.append(key)
            except Exception:
                pass
        for key in toDel:
            del complement[key]
        # read first line to identify datatype
        line1, line2, line3 = '', '', ''
        if not self.isFileContent:
            if not os.path.exists(self.filename):
                if not self.silent:
                    print('Class Graph: file do not exist!', self.filename)
                return False
            line1, line2, line3 = GraphIO.readDataFileLine123(self.filename)
        else:
            self.filename = self.filename.replace('\r', '\n')
            line1 = self.filename.split('\n')[0]
            if 'isfilecontent' in attributes:
                del attributes['isfilecontent']
            if attributes['label'] == '':
                del attributes['label']
        msg = ''
        # (uncomplete) list of classes handled:
        # with corresponding Curve types: CV, Cf, JV, TIV, MBE, MCA, SIMS
        # without: MCA fit (html output of XRF), MBElog
        loaded = False
        # first check if instructed to open the file in a certain way
        if ('readas' in attributes and
            attributes['readas'].lower() in ['database', 'generic']):
                msg = 'opened using standard methods.'
                GraphIO.readDataFromFileTxt(self, attributes)
                loaded = True
        # if not, then try every possibility
        if not loaded:
            childClasses = GraphIO._listGraphChildClasses()
    #        print(childClasses)
            for child in childClasses:
                if child.isFileReadable(fileNameBase, fileExt, line1=line1, line2=line2, line3=line3):
                    msg = 'opened as ' + child.FILEIO_GRAPHTYPE + '.'
                    self.headers.update({'meastype': child.FILEIO_GRAPHTYPE})
                    res = child.readDataFromFile(self, attributes, fileName=fileNameBase, line1=line1, line2=line2, line3=line3)
                    if res is None or res != False:
                        loaded = True
                        break
        # if not, then try default methods
        if not loaded:
            # other not readable formats
            if fileExt in ['.pdf']:
                msg = 'do not know how to open.'
            else:
                # keep the default opening mechanism in the main class
                msg = 'opened using standard methods.'
                GraphIO.readDataFromFileTxt(self, attributes)
        if msg != '':
            if not self.silent:
                print('File '+fileName.split('/')[-1]+'... ' + msg)
    
        
    def _funcReadDataFile(self, filecontent, attributes=None):
        """ Try to guess which method is best suited to open the file """
        if attributes is None:
            attributes = {}
        if 'readas' in attributes:
            if attributes['readas'].lower() == 'database':
                return GraphIO.readDataFromFileDatabase
            if attributes['readas'].lower() == 'generic':
                return GraphIO.readDataFromFileGeneric
        # test to identify if file is organized as series of headers + x-y columns, or as database
        for i in range(len(filecontent)):
            try:
                float(filecontent[i][0])
                # return fileGeneric only if the first eleemnt of any row can
                # interpreted as a number, ie. float(element) do not raise a
                # ValueError
                return GraphIO.readDataFromFileGeneric
            except ValueError:
                pass
        return GraphIO.readDataFromFileDatabase

        
    
    def readDataFromFileTxt(self, attributes):
        """
        Reads content of a .txt file, and parse it as a database or as a
        generic file.
        """
        if not self.isFileContent:
            try:
                fileContent = [resplit('[\t ,;]', line.strip(':\r\n')) for line in open(self.filename, 'r')]
            except UnicodeDecodeError as e:
                print('Exception', type(e), 'in GraphIO.readDataFromFileTxt:')
                print(e)
                return False
            except FileNotFoundError as e:
                return False
        else:
            fileContent = self.filename.split('\n')
            delimiter = GraphIO._readDataIdentifyDelimiter(self, fileContent)
            fileContent = [resplit(delimiter, line) for line in fileContent]
            while len(fileContent) > 0 and fileContent[-1] == ['']:
                del fileContent[-1]
        f = GraphIO._funcReadDataFile(self, fileContent, attributes)
        if not self.isFileContent:
            # intentionally do not provide file content, function
            # will read file again with the desired data processing
            f(self, attributes)
        else:
            self.filename = '.txt' # free some memory
            attributes.update({'filename': 'From clipboard'})
            f(self, attributes, fileContent=fileContent)
            self.isFileContent = False

            
    def _readDataIdentifyDelimiter(self, lines):
        # determine data character separator - only useful if reads file, otherwise is supposed to be already separated
        delimiter = '\t'
        nTab = np.sum([l.count('\t') for l in lines])
        threshold = np.ceil( max(1, len(lines)-5) / 2)
        if nTab < threshold: # if really not enough tabs in the file to be reasonable that data are tab-separated
            delimiters = [',', ';', ' ']
            for s in delimiters:
                n = np.sum([l.count(s) for l in lines])
                if n > threshold:
                    delimiter = s
                    print ('Graph file generic: data delimiter identified as "'+s+'".')
                    break
        return delimiter

        
    def readDataFromFileGeneric(self, attributes, fileContent=None,
                                ifReplaceCommaByPoint=False, **kwargs):
        """
        Reads the file as a column-organized data file, with some headers
        lines at the beginning.
        kwargs: are handled the following keywords:
        - delimiter: if provided, assumed the values is the text delimiter.
        - delimiterHeaders: same as delimiter, but only for the headers section
        """
        # idea:
        # 1/ read headers, store values in headers or in sampleInfo.
        # 2/ read data, identify data structure
        # 3/ look back at headers to fill in dataInfo according to data
        #    structure
        # if fileContent provided, file is not read and fileContent is used
        # instead. Supposed to be a 2-dimensional list [line][col]

        # update default information
        self.data = []
        parsedAttr = {}
        fileName, fileext = os.path.splitext(self.filename) # , fileExt=
        self.headers.update({'sample': fileName.split('/')[-1], 'meastype': FILEIO_GRAPHTYPE_UNDETERMINED})
        singlecurve= False
        if '_singlecurve' in attributes and attributes['_singlecurve']:
            singlecurve = True
            del attributes['_singlecurve']

        skipLines = 0
        skipFooters = 0
        lastSampleInfo = '' # key of last saved parameter
        lastLine = ''
        colLabels = []
        # load content of file
        if fileContent is None: # default behavior
            lines = [line.rstrip(':\r\n\t') for line in open(self.filename, 'r')]
            if ifReplaceCommaByPoint:
                lines = [resub('(?P<a>[0-9]),(?P<b>[0-9])', lambda word: word.group('a')+'.'+word.group('b'), line) for line in lines]
        else: # if some content was provided
            lines = fileContent
        # identify data separator (tab, semicolumn, etc.)
        if 'delimiter' in kwargs and isinstance(kwargs['delimiter'], str):
            delimiter = kwargs['delimiter']
        else:
            delimiter = GraphIO._readDataIdentifyDelimiter(self, lines)
        delimiterHeaders = None if 'delimiterHeaders' not in kwargs else kwargs['delimiterHeaders']
            
        
        def splitline(line, delimiter):
            couple = line.split(delimiter)
            return [couple[i].strip(' :') for i in range(len(couple))]
            
        # process header lines
        numEmpty = 0
        for line in lines:
            if fileContent is None:
                couple = splitline(line, delimiter)
            else:
                couple = line
            # stops looping at first line with numerical values
            if is_number(couple[0]):
                isLabelDefault = True if 'label' in attributes and 'filename' in attributes and attributes['label'] == '.'.join(attributes['filename'].split('\\')[-1].split('/')[-1].split('.')[:-1]).replace('_',' ') else False
                if len(lastLine) == 1:
                    couple2 = splitline(lastLine[0], delimiter)
                    if len(couple2) > 1:
                        lastLine = couple2
                if lastSampleInfo != '' and ('label' not in attributes or
                       (isLabelDefault and len(lastLine) > 1 and lastSampleInfo not in self.graphInfoKeys + self.headersKeys + self.dataInfoKeysGraph)):
                    # if some column names were identified, try to define labels by concatenating filename stripped from numerics with colum name
                    lbl = attributes['label'].split(' ') if 'label' in attributes else ['']
                    for i in range(len(lbl)-1, -1, -1):
                        try:
                            float(lbl[i])
                            del lbl[i]
                        except ValueError:
                            if lbl[i] in['File']:
                                del lbl[i]
                    suff = ' '.join(lbl)
                    if len(suff) > 1:
                        suff += ' '
                    colLabels = [suff + l.replace('"','') for l in lastLine]
                    if 'label' in attributes:
                        del attributes['label']
                # don't save last couple, intepreted as column name
                break
            # if not numerical value -> still information
            else:
                skipLines += 1
                # change delimiter is provided
                if delimiterHeaders is not None and fileContent is None:
                    couple = splitline(line, delimiterHeaders)
                # interpret data
                if couple[0] == '':
                    if len(couple) == 1 or len(''.join(couple)) == 0:
                        continue
                    couple[0] = 'empty'+str(numEmpty)
                    numEmpty+= 1
                val = couple[0]
                if len(couple) > 1:
                    val = float(couple[1]) if is_number(couple[1]) else couple[1]
                # all keywords in lowercase
                if not is_number(couple[0]):
                    couple[0] = couple[0].lower()
                
                # conversion from matlab terminology
                if couple[0] == 'figuresize':
                    couple[0] = 'figsize'
                    try:
                        if len(ast.literal_eval(val)) == 4:
                            val = ast.literal_eval(val)
                            val = val[2:4]
                            if min(val) > 30:
                                # assumes units are given in pixels -> convert in inches, assuming FILEIO_DPI screen resolution
                                val = [x / self.FILEIO_DPI for x in val]
                    except Exception:
                        print('GraphIO.readDataFromFileGeneric: Error when setting figure size')
                replace = {'facealpha': 'alpha'}
                if couple[0] in replace:
                    couple[0] = replace[couple[0]]
                # other conversion
                if couple[0] == 'legendlocation':
                    couple[0] = 'legendproperties'
                # end of matlab conversion

                # identify headers values
                if couple[0] in ['sampleInfo', 'meastype']:
                    self.headers[couple[0]] = val
                # identify graphInfo values
                elif couple[0] in ['text']:
                    self.graphInfo[couple[0]] = codecs.decode(val, 'unicode_escape')
                    tmp = stringToVariable(self.graphInfo[couple[0]].replace('\n','\\n').replace('\\t','\\\\t'))
                    if isinstance(tmp, list): # only if list, otherwise not want stringToVariable
                        self.graphInfo[couple[0]] = tmp
                elif couple[0] in ['legendproperties']:
                    if not isinstance(val, (int, float)):
                        val = stringToVariable(val)
                    self.graphInfo[couple[0]] = val
                elif couple[0] in self.graphInfoKeys or couple[0].startswith('subplots'): # standard reading for graphinfo parameters
                    self.graphInfo[couple[0]] = stringToVariable(val)
                # send everything else in Curve, leave nothing for sampleInfo
                else:
                    couple[0] = couple[0].replace('legend', 'label').replace('â²','2').replace('xycurve', 'curve')
                    if couple[0] in ['label']:
                        couple = [couple[0]] + [cpl.replace('\\n','\n').replace('\t','\\t') for cpl in couple[1:]]
                    parsedAttr.update({couple[0]: couple})
                
                if couple[0] in self.graphInfo:
                    if isinstance(self.graphInfo[couple[0]], str):
                        self.graphInfo[couple[0]] = self.graphInfo[couple[0]].replace('\\n', '\n')
                    elif isinstance(self.graphInfo[couple[0]], list):
                        for i in range(len(self.graphInfo[couple[0]])):
                            if isinstance(self.graphInfo[couple[0]][i], str):
                                self.graphInfo[couple[0]][i] = self.graphInfo[couple[0]][i].replace('\\n', '\n')

                lastSampleInfo = couple[0]
                lastLine = couple
        # check that text input are correct
        self.checkValidText()
        # if must parse last "header" line nevertheless (first element blank)
        if singlecurve and len(lastLine) > 1 and lastLine[0].startswith('empty'):
            linei = skipLines - 1
            # retrieve last header line and first content and compare if positions of numeric match
            couple1 = []
            if fileContent is None:
                couple0 = splitline(lines[linei], delimiter)
                if linei+1 < len(lines):
                    couple1 = splitline(lines[linei+1], delimiter)
            else:
                couple0 = lines[linei]
                if linei+1 < len(lines):
                    couple1 = lines[linei+1]
            if len(couple0) == len(couple1):
                isnum0 = [is_number(v) for v in couple0]
                isnum1 = [is_number(v) for v in couple1]
                test = [bool(isnum0[i] == isnum1[i]) for i in range(1, len(isnum0))]
                if np.array(test).all():
                    skipLines = max(skipLines-1, 0)
        
        # if last text line was not empty -> interpret that as x-y-labels
        if colLabels != []:
            self.headers['collabels'] = colLabels
        if 'collabels' not in self.headers:
            if 'label' in parsedAttr:
                self.headers['collabels'] = parsedAttr['label']
        if colLabels != []:
            if 'xlabel' not in self.graphInfo:
                self.graphInfo['xlabel'] = colLabels[0]
            if 'ylabel' not in self.graphInfo and len(colLabels) > 1:
                self.graphInfo['ylabel'] = colLabels[1]

        while lines[-1] is '':
            lines.pop()
            skipFooters += 1
        skipFooters = 0  # do not understand it... genfromtxt can fail when files ends with several \n and this is not set!?

        # look for data in file after headers
        if skipLines < len(lines):
            # default behavior
            if fileContent is None:
                usecols = range(0,len(lines[skipLines].split(delimiter)))
                dictConverters = {}
                if ifReplaceCommaByPoint:
                    lambd = lambda x: float(str(str(x, 'UTF-8')).replace(',','.'))
                    for i in usecols:
                        dictConverters.update({i: lambd})
                genFromTxtOpt = {}
                if len(dictConverters) > 0:
                    genFromTxtOpt.update({'converters': dictConverters})
                data = np.transpose(np.genfromtxt(self.filename, skip_header=skipLines, delimiter=delimiter, invalid_raise=False, usecols=usecols, skip_footer=skipFooters, **genFromTxtOpt))
            else:
                # if some content was provided
                for i in range(skipLines, len(fileContent)-skipFooters):
                    for j in range(len(fileContent[i])):
                        fileContent[i][j] = np.float64(fileContent[i][j]) if fileContent[i][j] != '' and is_number(fileContent[i][j]) else np.nan
                data = np.transpose(np.array(fileContent[skipLines:]))
            colX = 0
            cols = []
            
            # some checks, and build array test -> know which data column are not empty (returned as nan values)
            if len(data.shape) < 2 and len(lines) == skipLines + 1:
                # only 1 data row
                test = [np.isnan(v) for v in data]
                data = data.reshape((len(data), 1))
            else:
                if len(data.shape) < 2 and len(lines) > skipLines + 1:
                    # if only 1 data colum
                    data = np.array([range(len(data)), data])
                    for key in parsedAttr:
                        parsedAttr[key] = [parsedAttr[key][0]] + parsedAttr[key]
                test = [np.isnan(data[i,:]).all() for i in range(data.shape[0])]
            # if still cannot define colLabels
            if 'collabels' not in self.headers:
                self.headers['collabels'] = [''] * len(test)
                if len(data.shape) < 2:
                    self.headers['collabels'] = [''] * 2
            # do not fill in 'filename', or 'label' default value if filename is amongst the file parameters
            for key in parsedAttr:
                if key in attributes:
                    del attributes[key]
            if 'label' not in parsedAttr and 'collabels' in self.headers :
                parsedAttr['label'] = self.headers['collabels']
            # parse through data
            if singlecurve:
                self.data.append(Curve(data, attributes))
                cols.append(1)
            for colY in range(1, len(test)): # normal column guessing
                # if nan -> column empty -> start new xy pair
                if test[colY]:
                    colX = colY + 1
                elif colY > colX:
                    # removing of nan pairs is performed in the Curve constructor
                    self.data.append(Curve(np.append(data[colX,:], data[colY,:]).reshape((2, len(data[colY, :]))), attributes))
                    cols.append(colY)
            if len(cols) > 0:
                # correct parsedAttr to match data structure
                if len(self.headers['collabels']) <= max(cols):
                    self.headers['collabels'].extend([''] * max(cols))
                self.headers['collabels'] = [self.headers['collabels'][i] for i in cols]
                # apply parsedAttr to different curves
                for key in parsedAttr:
                    if len(parsedAttr[key]) < max(cols):
                        parsedAttr[key].extend([''] * max(cols))
                    for i in range(len(cols)):
                        try:
                            val = parsedAttr[key][cols[i]]
                        except IndexError:
                            continue
                        val = stringToVariable(val)
                        if val != '':
                            self.curve(i).update({key: val})
                # cast Curve in child class if required by parameter read
                for c in range(self.length()):
                    newtype = self.curve(c).getAttribute('curve')
                    if newtype not in ['', 'curve', 'curvexy']:
                        self.castCurve(newtype, c, silentSuccess=True)

                            
    def readDataFromFileDatabase(self, attributes, fileContent=None, **kwargs):
        """
        fileContent: list of lists containing file content (imagine csv file)
        kwargs: useless here
        """
        self.headers.update({'meastype': FILEIO_GRAPHTYPE_DATABASE})
        if fileContent is None:
            # this seems to parse the file satisfyingly
            fileContent = [line.strip(':\r\n').split('\t') for line in open(self.filename, 'r')]
        # suppress empty lines
        for i in np.arange(len(fileContent)-1, -1, -1):
            if sum([len(fileContent[i][j]) for j in range(len(fileContent[i]))]) == 0:
                del fileContent[i]
            else:
                for j in range(len(fileContent[i])):
                    fileContent[i][j] = fileContent[i][j].replace('\\n', '\n')
        # max length of a line
        fileNcols = max([len(fileContent[i]) for i in range(len(fileContent))])

        # start processing - identify column names
        self.headers.update({'collabelsdetail': [], 'rowlabels': [], 'rowlabelnums': []})
        # check if further lines as header - must be pure text lines.
        # we store lines consisting only in a pair of elements. In this case this should not be a column name, instead we will self.update()
        toUpdate = {}
        for i in range(0, len(fileContent)):
            line = fileContent[i]
            onlyText = True
            nVal = 0
            for val in line:
                if is_number(val):
                    onlyText = False
                    break
                if val != '':
                    nVal = nVal + 1
            if not onlyText:
                break

            if nVal == 1 and line[0] != '':
                split = line[0].split(': ')
                if len(split) > 1:
                    toUpdate.update({split[0]: split[1:]})
                else:
                    toUpdate.update({split[0]: ''})
            elif nVal == 2 and line[0] != '' and line[1] != '':
                toUpdate.update({line[0]: line[1]})
            else:
                self.headers['collabelsdetail'].append(fileContent[i][0:])
        # combine all column names identified so far
        nLinesHeaders = len(self.headers['collabelsdetail']) + len(toUpdate)
        labels = [''] * (fileNcols-1)
        for i in range(len(self.headers['collabelsdetail'])):
            for j in range(1, len(self.headers['collabelsdetail'][i])):
                labels[j-1] = labels[j-1] + '\n' + self.headers['collabelsdetail'][i][j]
        for i in range(len(labels)):
            labels[i] = labels[i][1:].strip('\n ')
        self.headers.update({'collabels': labels[:]}) # want a hard-copy here

        # parsing content
        if len(self.data) > 0:
            print('Class GraphIO.readDataFromDatabaseFile, data container NOT empty. Content of data deleted, previous data lost.')
        self.data = []
        # fill the Curve object with list of Curves, still empty at this stage
        for i in range(fileNcols - 1):
            self.data.append(Curve(np.zeros((2, len(fileContent)-nLinesHeaders)), attributes))
            self.data[-1].update({'label': labels[i]})

        # update the header info
        # first change datatype for display purposes 
        for key in toUpdate:
            if key in ['color']:
                toUpdate[key] = stringToVariable(toUpdate[key])
        self.update(toUpdate, ifAll=True)

        # loop over rows
        for i in range(len(fileContent)-nLinesHeaders):
            line = fileContent[i+nLinesHeaders]
            self.headers['rowlabels'].append(line[0])
            try:
#                self.headers['rowlabelnums'].append(float(''.join(ch for ch in line[0] if ch.isdigit() or ch=='.')))
                split = refindall('([0-9]*[.]*[0-9]*)', line[0])
                split = [s for s in split if len(s) > 0]
#                print (line[0], split)
                if len(split) > 0:
                    self.headers['rowlabelnums'].append(float(split[0]))
                else:
                    self.headers['rowlabelnums'].append(np.nan)
            except Exception:
                self.headers['rowlabelnums'].append(np.nan)
            for j in range(len(line)-1):
                self.data[j].setX(self.headers['rowlabelnums'][-1], index=i)
                if is_number(line[j+1]):
                    self.data[j].setY(float(line[j+1]), index=i)

                    
                    
    def filesave_default(self, filesave=''):
        """ (private) default filename when saving. """
        if filesave == '':
            fileName = self.filename.split('\\')[-1].split('/')[-1].replace('.', '_')
            filesave = fileName + '_'
            if self.length() > 0:
                filesave += self.curve(-1).getAttribute('complement')
            if filesave == '_':
                filesave = 'graphExport'
            if len(self.data) > 1:
                filesave = filesave + '_series'
        return filesave

    
    def export(self, filesave='', saveAltered=False, ifTemplate=False,
               ifCompact=True, ifOnlyLabels=False, ifClipboardExport=False):
        """
        Exports content of Grah object into a human- and machine- readable
        format.
        """
        if len(filesave) > 4 and filesave[-4:] == '.xml':
            return GraphIO.exportXML(self, filesave=filesave, saveAltered=saveAltered)
        # retrieve information of data alteration (plotting mode)
        alter = self.getAttribute('alter')
        if alter == '':
            alter = ['', '']
        if isinstance(alter, str):
            alter = ['', alter]
        # handle saving of original data or altered data
        tempAttr = {}
        if saveAltered: # if modified parameter, want to save modified data, and not save the alter attribute
            tempAttr = self.deleteAttr(['alter'])
        else:
            alter = ['', ''] # want to save original data, and alter atribute
        # preparation
        dataLe1 = 0
        out = ''
        # format general information for graph
        # template: do not wish any other information than graphInfo -> that looks good
        if not ifOnlyLabels: # if ifOnlyLabels: only export label argument
            for key in self.graphInfo:
                out = out + key + '\t' + str(self.graphInfo[key]).replace('\n', '\\n') + '\n'
            if not ifTemplate:
                keyListHeaders = ['meastype']
                for key in keyListHeaders:
                    if key in self.headers:
                        out = out + key + '\t' + str(self.headers[key]).replace('\n', '\\n') + '\n'
        # if compact mode, loop over curves to identify consecutive with identical x axis
        separator = ['\t'] + ['\t\t\t'] *  (self.length()-1) + ['\t']
        if ifCompact and not ifTemplate:
            for i in range(self.length()):
                if i > 0 and np.array_equal(self.curve(i).x(), self.curve(i-1).x()):
                    separator[i] = '\t'
        # format information for graph specific for each curve
        # first establish list of attributes, then loop over curves to construct export
        keysList = ['label']
        if not ifOnlyLabels: # if ifOnlyLabels: only export label argument
            for c in range(self.length()):
                for attr in self.curve(c).getAttributes():
                    if attr not in keysList:
                        #print ('Graph export', attr, keysList, attr in self.dataInfoKeysGraph, (not ifTemplate))
                        if attr in self.dataInfoKeysGraph or (not ifTemplate):
                            keysList.append(attr)
        # if save altered Curce export curves modified for alterations and offsets, and delte offset and muloffset key
        if saveAltered:
            funcx, funcy = Curve.x_offsets, Curve.y_offsets
            if 'offset' in keysList:
                keysList.remove('offset')
            if 'muloffset' in keysList:
                keysList.remove('muloffset')
        else:
            funcx, funcy = Curve.x, Curve.y
        # proceed with file writing
        keysList.sort()
        for key in keysList:
            out = out + key
            for c in range(self.length()):
                out = out + separator[c] + str(self.curve(c).getAttribute(key)).replace('\n', '\\n')
            out = out.rstrip('\t') + '\n'
        # format data
        for curve in self.data:
            if curve.shape(1) > dataLe1:
                dataLe1 = curve.shape(1)
        separator[0] = '\t\t\t'
        for i in range(dataLe1):
            for c in range(self.length()):
                if ifTemplate:
                    out += '1' + '\t' + '0' + '\t\t'
                else:
                    curve = self.curve(c)
                    if i < curve.shape(1):
                        if c > 0:
                            out += '\t'
                            if len(separator[c]) == 3:
                                out += '\t'
                        if len(separator[c]) == 3:
                            out += str(funcx(curve, index=i, alter=alter[0])) + '\t' + str(funcy(curve, index=i, alter=alter[1]))
                        else:
                            out += str(funcy(curve, index=i, alter=alter[1]))
                    else:
                        if c == 0:
                            separator[0] = '\t'
                        out += separator[c]
                            
            out += '\n'
            if ifTemplate:
                break
        # restore initial state of Graph object
        self.update(tempAttr)
        # end: return data, or save file
        if ifClipboardExport:
            return out
        else:
            # output
            filename = GraphIO.filesave_default(self, filesave) + '.txt'
            if self.getAttribute('saveSilent') != True:
                print('Graph data saved as', filename.replace('/','\\'))
            f = open(filename, 'w')
            f.write(out)
            f.close()
        return filename

    
    def exportXML(self, filesave='', **kwargs):
        """
        Exports Graph as .xml file
        NOT implemented:
            saveAltered=False, ifTemplate=False
        Non-sense here:
            ifCompact=True, ifClipboardExport=False
        """
        for key in kwargs:
            print('WARNING GraphIO.exportXML: received keyword', key, ':',
                  kwargs[key],', don\'t know what to do with it.')
        try:
            from lxml import etree
        except ImportError as e:
            print('Exception', type(e), 'in GraphIO.graphToXML:')
            print(e)
            return
        root = etree.Element("graph")
        # write headers
        for container in ['headers', 'sampleInfo', 'graphInfo']:
            if hasattr(self, container) and len(getattr(self, container)):
                tmp = etree.SubElement(root, container)
                for key in getattr(self, container):
                    tmp.set(key, str(getattr(self, container)[key]))
        # write curves
        for c in range(self.length()):
            curve = self.curve(c)
            tmp = etree.SubElement(root, 'curve'+str(c))
            attr = etree.SubElement(tmp, 'attributes')
            for key in curve.getAttributes():
                attr.set(key, str(curve.getAttribute(key)))
            data = etree.SubElement(tmp, 'data')
            x = etree.SubElement(data, 'x')
            y = etree.SubElement(data, 'y')
            x.text = ','.join([str(e) for e in curve.x()])
            y.text = ','.join([str(e) for e in curve.y()])
        # save
        filesave = GraphIO.filesave_default(self, filesave)
        if len(filesave) < 4 or filesave[-4:] != '.xml':
            filesave += '.xml'
        tree = etree.ElementTree(root)
        tree.write(filesave, pretty_print=True, xml_declaration=True, encoding="utf-8")
        if self.getAttribute('saveSilent') != True:
            print('Graph data saved as', filesave.replace('/','\\'))
        return filesave
        

    def _createSubplotsGridspec(self, figAx=None, ifSubPlot=False):
        """
        Generates a Figure (if not provided) with a gridpec axes matrix
        """
# TODO: Handle subplotskwargs: sharex True, 'col'
        import matplotlib.gridspec as gridspec
        # set fig as the active figure, if provided. Else with current figure
        if figAx is not None:
            fig, ax = figAx
            plt.figure(fig.number) # bring existing figure to front
            # delete all except the provided axis
            if not ifSubPlot:
                for ax_ in fig.get_axes():
                    if ax_ is not ax:
                        plt.delaxes(ax_)
        else:
            fig = plt.figure() # create a new figure
            ax = None
        # count number of graphs to be plotted
        axes = [{'ax': None, 'activenext': True} for c in self.iterCurves()]
        ncurvesondefault = 0 # if no curve is displayed on default axis, do not create it
        ngraphs = 0
        for curve in range(self.length()):
            c = self.curve(curve)
            if isinstance(c, Curve_Inset):
                axes[curve].update({'ax': 'inset', 'activenext': False})
                # default can be overridden if Graph file has length 0
            elif isinstance(c, Curve_Subplot):
                rowspan = int(c.getAttribute('subplotrowspan', 1))
                colspan = int(c.getAttribute('subplotcolspan', 1))
                ngraphs += rowspan * colspan
            elif ngraphs == 0:
                ncurvesondefault += 1
                ngraphs = 1
        ngraphs = max(1, ngraphs)
        # transpose?
        transpose = self.getAttribute('subplotstranspose', False)
        # determine axes matrix shape
        ncols = int(self.getAttribute('subplotsncols', (1 if ngraphs < 2 else 2)))
        nrows = int(np.ceil(ngraphs / ncols))
        # width, heigth ratios?
        gridspeckwargs = {}
        val = list(self.getAttribute('subplotswidth_ratios', ''))
        if len(val) > 0:
            target = ncols if not transpose else nrows
            if len(val) != target:
                val += [1] * max(0, (target)-len(val))
                while len(val) > target:
                    del val[-1]
                print('GraphIO._createSubplotsGridspec: corrected width_ratios to match ncols', ncols, ':', val)
            gridspeckwargs.update({'width_ratios' : val})
        val = list(self.getAttribute('subplotsheight_ratios', ''))
        if len(val) > 0:
            target = nrows if not transpose else ncols
            if len(val) != target:
                val += [1] * max(0, (target)-len(val))
                while len(val) > target:
                    del val[-1]
                print('GraphIO._createSubplotsGridspec: corrected height_ratios to match nrows', nrows, ':', val)
            gridspeckwargs.update({'height_ratios': val})
        # generate axes matrix: either gs, either ax is created
        gs, matrix = None, np.array([[]])
        if ax is None:
            if ngraphs == 1:
                ax = fig.add_subplot(111)
                ax.ticklabel_format(useOffset=False)
                ax.patch.set_alpha(0.0)
                # hide axis if no Curve will be plotted on it
                if ncurvesondefault == 0:
                    ax.axis('off')
            else:
                matrix = np.ones((nrows, ncols)) * (-1)
                if transpose:
                    gs = gridspec.GridSpec(ncols, nrows, **gridspeckwargs)
                else:
                    gs = gridspec.GridSpec(nrows, ncols, **gridspeckwargs)
        
        # coordinates of the plot id
        subplotsid = self.getAttribute('subplotsid', False)
        if (subplotsid is not False and
            (not isinstance(subplotsid, (list, tuple)) or len(subplotsid) != 2)):
            subplotsid = (-0.03, 0.00)
        # misc adjustments to self
        if ngraphs > 1: # do not want default values if multiple subplots
            self.update({'xlabel': '', 'ylabel': ''})
        # return
        return [fig, ax], gs, matrix, subplotsid, ngraphs

        

    def _newaxis(self, curve, fig, gs, matrix, subplotsid, subplotscounter, subplotidkwargs):
        rowspan = int(curve.getAttribute('subplotrowspan', 1))
        colspan = int(curve.getAttribute('subplotcolspan', 1))
        gspos = gs.get_grid_positions(fig)
        flag = False
        for i_ in range(len(matrix)):
            for j_ in range(len(matrix[i_])):
                if matrix[i_,j_] == -1:
                    # first free spot found: start there
                    if rowspan > matrix.shape[0] - i_:
                        # calculation of number of rows probably faulty
                        print('WARNING GraphIO.plot: rowspan larger than can handled for subplot', subplotscounter, ', value coerced to', matrix.shape[0] - i_)
                    if colspan > matrix.shape[1] - j_:
                        # either faulty calculation of number of rows,
                        # either that colspan cannot fit in ncols
                        print('WARNING GraphIO.plot: colspan larger than can handled for subplot', subplotscounter, ', value coerced to', matrix.shape[1] - j_)
                    rowspan = min(rowspan, matrix.shape[0] - i_)
                    colspan = min(colspan, matrix.shape[1] - j_)
                    if self.getAttribute('subplotstranspose', False):
                        if subplotsid:
                            coords = (gspos[2][i_]+subplotsid[0], gspos[1][j_]+subplotsid[1])
                            self.addText('('+chr(ord('a')+subplotscounter)+')', coords, subplotidkwargs)
                        ax = plt.subplot(gs[j_:(j_+colspan), i_:(i_+rowspan)])
                    else:
                        if subplotsid:
                            coords = (gspos[2][j_]+subplotsid[0], gspos[1][i_]+subplotsid[1])
                            self.addText('('+chr(ord('a')+subplotscounter)+')', coords, subplotidkwargs)
                        ax = plt.subplot(gs[i_:(i_+rowspan), j_:(j_+colspan)])
                    matrix[i_:(i_+rowspan), j_:(j_+colspan)] = subplotscounter
                    flag = True
                    break
            if flag:
                break
        if ax is None:
            print('ERROR GraphIO.plot create subplot: ax is None')
            print('matrix', matrix)
            raise Exception
        ax.ticklabel_format(useOffset=False)
        ax.patch.set_alpha(0.0)
        return ax

        
        
    # creat graph from data with headers additional data
    def plot(self, filesave='', imgFormat='', figsize=(0, 0), ifSave=True, ifExport=True, figAx=None, ifSubPlot=False, handles=None):
        """
        Plot the content of the object.
        imgFormat: by default image will be .png. Possible format are the ones
            accepted by plt.savefig
        figsize: default figure size is a class consant
        filesave: (optional) filename for the saved graph
        ifExport: if True, reate a human- and machine readable .txt file
            containing all information of the graph
        figAx= [fig, ax] provided. Useful when wish to embedd graph in a GUI.
        ifSubPlot: False, unless fig to be drawn in a subplot. Not handled by
            the GUI, but might be useful in scripts. Prevents deletion of
            existing axes.
        """
        # treat filesave, and store info if not already done. required for
        # relative path to subplots or insets
        filename = GraphIO.filesave_default(self, filesave)
        if not hasattr(self, 'filename'):
            self.filename = filename

        # store some attributes which might be modified upon execution
        restore = {}
        for attr in ['text', 'textxy', 'textargs']:
            restore.update({attr: self.getAttribute(attr)})
        
        # retrieve default axis positions subplotAdjustDefault
        subplotAdjustDefault = {}
        for key in ['bottom', 'left', 'right', 'top', 'hspace', 'wspace']:
            subplotAdjustDefault.update({key: mpl.rcParams['figure.subplot.'+key]})
        subplotColorbar = [0.90, subplotAdjustDefault['bottom'], 0.05, subplotAdjustDefault['top'] - subplotAdjustDefault['bottom']]
        # shift fdefault right if there is a colorbar defined in a Curve
        for c in self.iterCurves():
            if c.getAttribute('colorbar') is not '':
                subplotAdjustDefault['right'] -= 0.1
                break

        # other data we want to retrieve now
        fontsize = Graph.DEFAULT['fontsize'] if 'fontsize' not in self.graphInfo else self.graphInfo['fontsize']
        alter = self._getAlter()
        ignoreXLim = True if alter[0] != '' else False
        ignoreYLim = True if alter[1] != '' else False
        axTwinX, axTwinY, axTwinXY = None, None, None
        
        # check
        if len(self.data) <= 0:
            if ifSubPlot:
                pass
            else:
                print('Warning GraphIO.plot', self.filename, ': no data to plot!')
                if figAx is not None:
                    fig, ax = figAx[0], figAx[1]
                    if ax is not None:
                        plt.sca(ax)
                    plt.cla()
                #return 1

        # retrieve figure size
        if figsize == (0, 0): # if figure size not imposed at function call, look for instructions in measurement data
            if 'figsize' in self.graphInfo:
                figsize = self.graphInfo['figsize']
            else: 
                figsize = self.FIGSIZE_DEFAULT
        
        # create graph, initialize axes
        subplotscounter = 0
        [fig, ax], gs, matrix, subplotsid, subplotsngraphs = GraphIO._createSubplotsGridspec(self, figAx, ifSubPlot=ifSubPlot)
        # either ax is None and gs is not None, either the other way around
        subplotsncols, subplotsnrows = matrix.shape
        subplotidkwargs = {'textcoords': 'figure fraction',
                           'fontsize': fontsize,
                           'horizontalalignment': 'right',
                           'verticalalignment': 'top'}
        axes = []
        if ax is not None:
            axes.append(ax)
        fig.patch.set_alpha(0.0)
        if not ifSubPlot: # set figure size if fig is not a subplot
            fig.set_size_inches(*figsize, forward=True)
        colorbar_ax = []
        
        # adjust positions of the axis within the graph (subplots_adjust)
        subplotAdjust = dict(subplotAdjustDefault)
        if 'subplots_adjust' in self.graphInfo:
            val = deepcopy(self.graphInfo['subplots_adjust'])
            if isinstance(val, dict):
                subplotAdjust.update(val)
            elif isinstance(val, list):
                # handles relative in given in absolute
                if isinstance(val[-1], str) and val[-1] == 'abs':
                    del val[-1]
                    fs = fig.get_size_inches()
                    for i in range(len(val)):
                        val[i] = val[i] / fs[i % 2]
                indic = ['left', 'bottom', 'right', 'top', 'wspace', 'hspace']
                for i in range(min(len(val), len(indic))):
                    subplotAdjust.update({indic[i]: val[i]})
            else:
                subplotAdjust.update({'bottom': val})
        if fig is not None:
            fig.subplots_adjust(**subplotAdjust)
        if gs is not None:
            gs.update(**subplotAdjust)
        # auxiliary graph, to be created when new axis, filled with curves and plotted when new axis is created
        graphAux = None # auxiliary Graph, calling plot() when creating new axis
        graphAuxKw = {'config': self._config.filename}
        
        # plot data
        type_plot = self.getAttribute('typeplot')
        # set default graph scales (lin, log, etc.)
        if type_plot != '' and gs is None:
            xarg = 'log' if type_plot in ['semilogx', 'loglog'] else 'linear'
            try:
                ax.set_xscale(xarg)
            except (ValueError, AttributeError):
                print('Display error, try to go back to linear Transform ?')
            yarg = 'log' if type_plot in ['semilogy', 'loglog'] else 'linear'
            try:
                ax.set_yscale(yarg)
            except (ValueError, AttributeError):
                print('Display error, try to go back to linear Transform?')
        
        boxplot    = {'y':[], 'positions':[], 'labels':[], 'color':[], 'i':0}
        violinplot = {'y':[], 'positions':[], 'labels':[], 'color':[]}
        violinplotkwargs = {}
        
        ignoreNext = 0
        if handles is None or not isinstance(handles, list):
            handles = []
        # Start looping on the curves
        for curve_i in range(self.length()):
            handle = None
            if ignoreNext > 0:
                ignoreNext -= 1
                continue # ignore this curve, go the next one...
                # for example if previous curve type was scatter
            curve = self.curve(curve_i)
            attr = curve.getAttributes()
            # if curve is hidden: ignore it
            if curve.isHidden():
                continue
            
            # Inset: if curve contains information for an inset in the Graph?
            if isinstance(curve, Curve_Inset):
                val = curve.getAttribute('insetfile')
                inset = Graph(self.filenamewithpath(val), **graphAuxKw)
                coords = list(attr['insetcoords']) if 'insetcoords' in attr else [0.3,0.2,0.4,0.3]
                if 'insetupdate' in attr and isinstance(attr['insetupdate'], dict):
                    inset.update(attr['insetupdate'])
                axInset = fig.add_axes(coords)
                axInset.ticklabel_format(useOffset=False)
                axInset.patch.set_alpha(0.0)
                if curve.getAttribute('insetfile') in['', ' '] or inset.length() == 0:
                    # nothing in provided Graph -> created axis becomes active one
                    if graphAux is not None:
                        # if there was already auxiliary graph: display it, create anew
                        GraphIO.plot(graphAux, ifSave=False, ifExport=False,
                                     figAx=[fig, ax], ifSubPlot=True)
                    graphAux = inset
                    ax = axInset
                    plt.sca(ax)
                    axTwinX, axTwinY, axTwinXY = None, None, None
                else:
                    # found a Graph, place it in inset. Next Curve in existing
                    # axes. No change for graphAux
                    inset.plot(figAx=[fig, axInset], ifSave=False, ifExport=False,
                               ifSubPlot=True)
                    continue # go to next Curve
            
            # Subplots: if more than 1 subplot is expected 
            if gs is not None:
                # if required, create the new axis
                if ax is None or isinstance(curve, Curve_Subplot):
                    if graphAux is not None:
                        GraphIO.plot(graphAux, ifSave=False, ifExport=False,
                                     figAx=[fig, ax], ifSubPlot=True)
                        graphAux = None
                    ax = GraphIO._newaxis(self, curve, fig, gs, matrix, subplotsid, subplotscounter, subplotidkwargs)
                    plt.sca(ax)
                    axTwinX, axTwinY, axTwinXY = None, None, None
                    subplotscounter += 1
                    axes.append(ax)
                else: # we go on, not enough information to create subplot
                    pass
            # shall we plot a Graph object instead of a Curve in this new axis?
            if isinstance(curve, Curve_Subplot):
                # new axes, so create new graphAux (is None by construction)
                val = curve.getAttribute('subplotfile')
                if val not in [' ', '', None]:
                    graphAux = Graph(self.filenamewithpath(val), **graphAuxKw)
                else:
                    graphAux = Graph('', **graphAuxKw)
                upd = {'subplots_adjust': ''}
                if isinstance(curve.getAttribute('subplotupdate'), dict):
                    upd.update(curve.getAttribute('subplotupdate'))
                graphAux.update(upd)
                if val not in [' ', '', None]:
                    continue # go to next Curve
                # if no file then try to plot data of the Curve
            
            if graphAux is not None:
                # if active axis is not main one, place Curve in graphAux
                if isinstance(curve, (Curve_Inset, Curve_Subplot)):
                    graphAux.append(Curve(curve.data, curve.getAttributes()))
                else:
                    graphAux.append(curve)
                continue
            
            # twin axis: look which axis to use
            ax_ = ax
            ifTwinX = (curve.getAttribute('ax_twinx') in [True, 1])
            ifTwinY = (curve.getAttribute('ax_twiny') in [True, 1])
            if ifTwinX and ifTwinY:
                if axTwinXY is None:
                    axTwinXY = ax.twinx().twiny()
                ax_ = axTwinXY
            elif ifTwinX:
                if axTwinX is None:
                    axTwinX = ax.twinx()
                ax_ = axTwinX
            elif ifTwinY:
                if axTwinY is None:
                    axTwinY = ax.twiny()
                ax_ = axTwinY
            
            # do the actual plotting
            try:
                handle, ignoreNext = curve.plot(ax_, graph=self, graph_i=curve_i,
                        type_plot=type_plot, ignoreNext=ignoreNext,
                        boxplot=boxplot,
                        violinplot=violinplot, violinplotkwargs=violinplotkwargs)
            except Exception as e:
                # need to catch exceptions as we want to proceed with graph generation
                print(type(e).__name__, 'Exception occured in Curve.plot function, Curve no', curve_i,'. Error message:')
                print(e)
                print(sys.exc_info()[0].__name__, sys.exc_info()[-1].tb_lineno)
            
            # Add colorbar if required
            if handle is not None and curve.getAttribute('colorbar', False):
                kwargs = {}
                if isinstance(attr['colorbar'], dict):
                    kwargs = deepcopy(attr['colorbar'])
                adjust = deepcopy(subplotColorbar)
                if 'adjust' in kwargs:
                    adjust = kwargs['adjust']
                    del kwargs['adjust']
                if isinstance(adjust[-1], str):
                    if adjust[-1] == 'ax':
                        # coordinates relative to the ax and not to figure
                        axbounds = ax_.get_position().bounds
                        adjust[0] = adjust[0] * axbounds[2] + axbounds[0]
                        adjust[1] = adjust[1] * axbounds[3] + axbounds[1]
                        adjust[2] = adjust[2] * axbounds[2]
                        adjust[3] = adjust[3] * axbounds[3]
                    else:
                        print('WARNING GraphOI.plot colorbar: cannot interpret last keyword (must be numeric, or \'ax\')')
                    del adjust[-1]
                modifier = []
                if 'labelsize' in kwargs:
                    pass
#                    modifier.append()
                    # TODO: 
                colorbar_ax.append({'ax': fig.add_axes(adjust), 'adjusted': False})
                colorbar_ax[-1]['cbar'] = fig.colorbar(handle, cax=colorbar_ax[-1]['ax'], **kwargs)
                try:
                    colorbar_ax[-1]['cbar'].solids.set_rasterized(True)
                    colorbar_ax[-1]['cbar'].solids.set_edgecolor('face')
                except Exception as e:
                    pass
                subplotColorbar[0] -= 0.11 # default for next colorbar
                # set again ax_ as the current plt active ax, not the colorbar axis
                plt.sca(ax_)

            # store handle in list
            if handle is not None:
                if isinstance(handle, list):
                    for h in handle:
                        handles.append({'handle': h})
                else:
                    handles.append({'handle': handle})
                    if curve.getAttribute('type') == 'scatter':
                        handles[-1].update({'setlegendcolormap': True})
        # end of loop over Curves

        # generates boxplot and violinplot, somehow differently than others
        if len(boxplot['y']) > 0:
            boxplot_dict = ax.boxplot(boxplot['y'], positions=boxplot['positions'], labels=boxplot['labels'])
            nb_el_boxplot = {'whiskers': 2, 'caps': 2}
            for key in boxplot_dict:
                for i in range(len(boxplot_dict[key])):
                    j = i if key not in nb_el_boxplot else int(np.floor(i / nb_el_boxplot[key]))
                    if j < len(boxplot['color']) and boxplot['color'][j] != '':
                        boxplot_dict[key][i].set_color(boxplot['color'][j])
        if len(violinplot['y']) > 0:
            violinplot_dict = ax.violinplot(violinplot['y'], positions=violinplot['positions'], **violinplotkwargs)#, labels=violinplot['labels'])
            for key in violinplot_dict:
                if isinstance(violinplot_dict[key], list):
                    for i in range(len(violinplot_dict[key])):
                        if violinplot['color'][i] != '':
                            violinplot_dict[key][i].set_color(violinplot['color'][i])
                else:
                    violinplot_dict[key].set_color([0,0,0])
            # violinplots labels not set automatically like boxplot
            for i in range(len(violinplot['labels'])):
                xticks = list(ax_.get_xticks())
                xtklbl = list(ax_.get_xticklabels())
                xticks.append(violinplot['positions'][i])
                xtklbl.append(violinplot['labels'][i])
                ax_.set_xticks     (xticks)
                ax_.set_xticklabels(xtklbl)
        
        # final display of auxiliary axes/graph
        if graphAux is not None:
            GraphIO.plot(graphAux, ifSave=False, ifExport=False,
                         figAx=[fig, ax], ifSubPlot=True)
            graphAux = None
            # main graph: work on initial axis
            ax = axes[0]
            plt.sca(ax)
            
        # usual graph cosmetics
        curvedummy = self.curve(-1) if self.length() > 0 else Curve([[0],[0]],{})
        fontsizeset = []
        def setAxisLabel(method, label):
            out = {'size': False}
            if isinstance(label, list) and len(label) == 2 and isinstance(label[1], dict):
                method(label[0], **label[1])
                if 'size' in label[1] or 'fontsize' in label[1]:
                    out['size'] = True
            else:
                method(label)
            return out
        if 'title' in self.graphInfo:
            out = setAxisLabel(ax.set_title, self.graphInfo['title'])
            if out['size']:
                fontsizeset.append(ax.title)
        if 'xlabel' in self.graphInfo:
            out = setAxisLabel(ax.set_xlabel, self.graphInfo['xlabel'])
            if out['size']:
                fontsizeset.append(ax.xaxis.label)
        if 'ylabel' in self.graphInfo:
            out = setAxisLabel(ax.set_ylabel, self.graphInfo['ylabel'])
            if out['size']:
                fontsizeset.append(ax.yaxis.label)
        # labels twin axis
        if 'twinx_ylabel' in self.graphInfo:
            if axTwinX is not None:
                out = setAxisLabel(axTwinX.set_ylabel, self.graphInfo['twinx_ylabel'])
                if out['size']:
                    fontsizeset.append(axTwinX.yaxis.label)
            elif axTwinXY is not None:
                out = setAxisLabel(axTwinXY.set_ylabel, self.graphInfo['twinx_ylabel'])
                if out['size']:
                    fontsizeset.append(axTwinXY.yaxis.label)
        if 'twiny_xlabel' in self.graphInfo:
            if axTwinY is not None:
                out = setAxisLabel(axTwinY.set_xlabel, self.graphInfo['twiny_xlabel'])
                if out['size']:
                    fontsizeset.append(axTwinY.xaxis.label)
            elif axTwinXY is not None:
                out = setAxisLabel(axTwinXY.set_xlabel, self.graphInfo['twiny_xlabel'])
                if out['size']:
                    fontsizeset.append(axTwinXY.xaxis.label)
        # lcoation of labels
        if 'xlabel_coords' in self.graphInfo:
            val = self.graphInfo['xlabel_coords']
            if isinstance(val, list):
                ax.xaxis.set_label_coords(val[0], val[1])
            else:
                ax.xaxis.set_label_coords(0.5, val)
        if 'ylabel_coords' in self.graphInfo:
            val = self.graphInfo['ylabel_coords']
            if isinstance(val, list):
                ax.yaxis.set_label_coords(val[0], val[1])
            else:
                ax.yaxis.set_label_coords(val, 0.5)
        # other
        if 'axhline' in self.graphInfo:
            if isinstance(self.graphInfo['axhline'], list):
                for h in self.graphInfo['axhline']:
                    pos = curvedummy.y(alter=alter[1], xyValue=[1, h], errorIfxyMix=True, neutral=True)
                    if not pos == 0 or not type_plot in ['semilogy', 'loglog']:
                        ax.axhline(pos, 0, 1, color='k')
            else:
                pos = curvedummy.y(alter=alter[1], xyValue=[1, self.graphInfo['axhline']], errorIfxyMix=True, neutral=True)
                if not pos == 0 or not type_plot in ['semilogy', 'loglog']:
                    ax.axhline(pos, 0, 1, color='k')
        if 'axvline' in self.graphInfo:
            if isinstance(self.graphInfo['axvline'], list):
                for v in self.graphInfo['axvline']:
                    pos = curvedummy.x(alter=alter[0], xyValue=[v, 1], errorIfxyMix=True, neutral=True)
                    if not pos == 0 or not type_plot in ['semilogx', 'loglog']:
                        ax.axvline(pos, 0, 1, color='k')
            else:
                pos = curvedummy.x(alter=alter[0], xyValue=[self.graphInfo['axvline'], 1], errorIfxyMix=True, neutral=True)
                if not pos == 0 or not type_plot in ['semilogx', 'loglog']:
                    ax.axvline(pos, 0, 1, color='k')
        
        # xlim, ylim. Start with xtickslabels as this guy would override xlim
        if 'xtickslabels' in self.graphInfo and not ignoreXLim:
            val = self.graphInfo['xtickslabels']
            if not isinstance(val, list):
                val = [val, [], {}]
            ticksloc = val[0]
            if not isinstance(ticksloc, list):
                ticksloc = ax.xaxis.get_ticklocs()
            ar = [val[1]] if len(val) > 1 and val[1] is not None else []
#                ticksvals = [a.get_text() for a in ax.xaxis.get_ticklabels()] # don't need to retrieve if do not change them
            kw = val[2] if len(val) > 2 and isinstance(val[2], dict) else {}
            if 'size' in kw:
                fontsizeset.append('ax.get_xticklabels()')
            plt.xticks(ticksloc, *ar, **kw)
        if 'ytickslabels' in self.graphInfo and not ignoreYLim:
            val = self.graphInfo['ytickslabels']
            if not isinstance(val, list):
                val = [val, [], {}]
            ticksloc = val[0]
            if not isinstance(ticksloc, list):
                ticksloc = ax.yaxis.get_ticklocs()
            ar = [val[1]] if len(val) > 1 and val[1] is not None else []
#                ticksvals = [a.get_text() for a in ax.yaxis.get_ticklabels()] # don't need to retrieve if do not change them
            kw = val[2] if len(val) > 2 and isinstance(val[2], dict) else {}
            if 'size' in kw:
                fontsizeset.append('ax.get_yticklabels()')
            plt.yticks(ticksloc, *ar, **kw)
        def alterLim(ax, lim, xory):
            limAuto = ax.get_xlim() if xory == 'x' else ax.get_ylim()
            lim = [l if not isinstance(l, str) else np.inf for l in lim]
            try:
                if xory == 'x':
                    fun = ax.set_xlim
                    lim = list(curvedummy.x(alter=alter[0], xyValue=[lim, [0]*len(lim)], errorIfxyMix=True, neutral=True))
                elif xory == 'y':
                    fun = ax.set_ylim
                    lim = list(curvedummy.y(alter=alter[1], xyValue=[[0]*len(lim), lim], errorIfxyMix=True, neutral=True))
                else:
                    return False
            except ValueError:
                return False
            except TypeError:
                lim = limAuto
            lim = [limAuto[i] if np.isnan(lim[i]) or np.isinf(lim[i]) else lim[i] for i in range(len(lim))]
            fun((lim[0], lim[1]))
        # xlim should be after xtickslabels
        if 'xlim' in self.graphInfo:
            alterLim(ax, deepcopy(self.graphInfo['xlim']), 'x')
        if 'ylim' in self.graphInfo:
            alterLim(ax, deepcopy(self.graphInfo['ylim']), 'y')
        if 'twinx_ylim' in self.graphInfo:
            if axTwinXY is not None:
                alterLim(axTwinXY, deepcopy(self.graphInfo['twinx_ylim']), 'y')
            if axTwinX is not None:
                alterLim(axTwinX , deepcopy(self.graphInfo['twinx_ylim']), 'y')
        if 'twiny_xlim' in self.graphInfo:
            if axTwinXY is not None:
                alterLim(axTwinXY, deepcopy(self.graphInfo['twiny_xlim']), 'y')
            if axTwinY is not None:
                alterLim(axTwinY , deepcopy(self.graphInfo['twiny_xlim']), 'y')
        if 'xticksstep' in self.graphInfo and not ignoreXLim:
            val = self.graphInfo['xticksstep']
            if isinstance(val, (list)):
                arr = val
            else:
                val = abs(val)
                axlim = ax.get_xlim()
                ticks = [t for t in ax.xaxis.get_ticklocs() if t >= min(axlim) and t <= max(axlim)]
                start, end = min(ticks), max(ticks)
                while start - val >= min(axlim):
                    start -= val
                while end + val <= max(axlim):
                    end += val
#                start, end = ax.get_xlim()
                arr = np.arange(start, end+end/1e10, val)
            ax.xaxis.set_ticks(arr)
        if 'yticksstep' in self.graphInfo and not ignoreYLim:
            val = self.graphInfo['yticksstep']
            if isinstance(val, (list)):
                arr = val
            else: #start, end = ax.get_ylim()
                val = abs(val)
                axlim = ax.get_ylim()
                ticks = [t for t in ax.yaxis.get_ticklocs() if t >= min(axlim) and t <= max(axlim)]
                start, end = min(ticks), max(ticks)
                while start - val >= min(axlim):
                    start -= val
                while end + val <= max(axlim):
                    end += val
                arr = np.arange(start, end+end/1e10, self.graphInfo['yticksstep'])
            ax.yaxis.set_ticks(arr)

        
        # text annotations
        if 'text' in self.graphInfo:
            textxyDefault = (0.05, 0.95)
            text = deepcopy(self.getAttribute('text'))
            texy = deepcopy(self.getAttribute('textxy', None))
            args = deepcopy(self.getAttribute('textargs'))
            if not isinstance(text, list):
                text = [text]
            if not isinstance(texy, list):
                texy = [texy] * len(text)
            else: # can be for example [0.05,0.95] -> should be duplicated
                if len(texy) == 2 and not isinstance(texy[0], (list, tuple)) and not isinstance(texy[1], (list, tuple)):
                    if texy == ['', '']:
                        texy = [None] * len(text)
                    else:
                        texy = [texy] * len(text)
            while len(texy) < len(text): # duplicate if missing elements
                texy.append(texy[-1])
            if not isinstance(args, list):
                args = [args]
            args = [deepcopy(a) for a in args] # might be dplicate by construction
            while len(args) < len(text):
                args.append(deepcopy(args[-1]))
            for i in range(len(text)):
                if text[i] != '':
                    if texy[i] != '' and texy[i] is not None:
                        if 'xytext' in args[i]:
                            print('Graph plot annotate:', text[i], 'textxy', texy[i], 'override', args[i]['xytext'])
                        args[i].update({'xytext': texy[i]})
                    if 'xytext' not in args[i]:
                        if 'xy' in args[i]:
                            args[i].update({'xytext': args[i]['xy']})
                        else:
                            args[i].update({'xytext': textxyDefault})
                    if 'xy' not in args[i]:
                        args[i].update({'xy': args[i]['xytext']}) # this one exists for sure
                    if 'xycoords' not in args[i] and 'textcoords' in args[i]:
                        args[i].update({'xycoords': args[i]['textcoords']})
                    if 'xycoords' in args[i] and 'textcoords' not in args[i]:
                        args[i].update({'textcoords': args[i]['xycoords']})
                    if 'textcoords' not in args[i]:
                        args[i].update({'textcoords': 'figure fraction'})
                    if 'xycoords' not in args[i]:
                        args[i].update({'xycoords': 'figure fraction'})
                    if 'fontsize' not in args[i] and 'fontsize' in self.graphInfo:
                        args[i].update({'fontsize': self.graphInfo['fontsize']})
                    #print('Graph plot annotate', text[i], 'args', args[i])
                    ax.annotate(text[i], **args[i])
        # legend
        if gs is None: # create legend on current ax
            legPropUser = deepcopy(self.getAttribute('legendproperties', default='best'))
            if not isinstance(legPropUser, dict):
                legPropUser = {'loc': legPropUser}
            if 'loc' in legPropUser:
                legPropUser['loc'] = str(legPropUser['loc']).lower()
                rep = {'best': 0, 'ne': 1, 'nw': 2, 'sw': 3, 'se': 4, 'right': 5, 'e': 6, 'w': 7, 's': 8, 'n': 9, 'center': 10}
                if legPropUser['loc'] in rep:
                    legPropUser['loc'] = rep[legPropUser['loc']]
            prop = {} if 'fontsize' not in self.graphInfo else {'size': self.graphInfo['fontsize']}
            if 'prop' in legPropUser:
                prop.update(legPropUser['prop'])
            legPropUser['prop'] = prop
            if 'fontsize' in legPropUser:
                legPropUser['prop'].update({'size': legPropUser['fontsize']})
                del legPropUser['fontsize']
            legLabelColor = None
            if 'color' in legPropUser:
                legLabelColor = legPropUser['color'].lower()
                del legPropUser['color']
            legProp = {'fancybox': True, 'framealpha': 0.5, 'frameon': False, 'numpoints': 1, 'scatterpoints': 1}
            legProp.update(legPropUser)
            labels = []
            for i in range(len(handles)-1, -1, -1):
                label = handles[i]['handle'].get_label() if hasattr(handles[i]['handle'], 'get_label') else '' # not legend if dont know how to find it
                if isinstance(handles[i]['handle'], mpl.image.AxesImage):
                    label = None
                if label is None or len(label) == 0 or label[0] == '_':
                    del handles[i] # delete curve handle if no label is to be shown
                else:
                    labels.append(label)
            leg = ax.legend([h['handle'] for h in handles], labels[::-1], **legProp)# labels is reversed by construction
            # color of legend
            if legLabelColor is not None:
                if legLabelColor == 'curve': # special value: same color for text and lines
                    lines, texts = leg.get_lines(), leg.get_texts()
                    if len(texts) > len(lines): # issue with errorbar, notably
                        lines = []
                        for h in handles:
                            if isinstance(h['handle'], tuple):
                                lines.append(h['handle'][0])
                            else:
                                lines.append(h['handle'])
                    for line, text in zip(lines, texts):
                        try:
                            text.set_color(line.get_color())
                        except:
                            pass
                else:
                    for text in leg.get_texts():
                        text.set_color(legLabelColor)
            # legend title
            legTitle = deepcopy(self.getAttribute('legendtitle')) # legendTitle can be of the form ['some title', {'fontsize':24}]
            if legTitle != '':
                setfunc = []
                if isinstance(legTitle, list):
                    for key in ['color', 'position', 'fontsize', 'align']:
                        if key in legTitle[1]:
                            setfunc.append([key, legTitle[1][key]])
                            del legTitle[1][key]
                    prop.update(legTitle[1])
                    legTitle = legTitle[0]
                leg.set_title(legTitle, prop=prop)
                for setf in setfunc:
                    if setf[0] == 'align':
                        leg._legend_box.align = setf[1]
                    else:
                        if hasattr(leg.get_title(), 'set_'+setf[0]):
                            getattr(leg.get_title(), 'set_'+setf[0])(setf[1])
                        else:
                            print('Warning GraphIO do not knwo what to do with keyword',setf[0])
            # in legend, set color of scatter correctly
            for i in range(len(handles)):
                if 'setlegendcolormap' in handles[i] and handles[i]['setlegendcolormap']:
                    try:
                        leg.legendHandles[i].set_color(handles[i]['handle'].get_cmap()(.5))#plt.cm.afmhot(.5))
                    except AttributeError as e:
                        print('error setlegendcolormap', e)
                        pass
                
        # arbitrary functions
        if 'arbitraryfunctions' in self.graphInfo:
            for fun in self.graphInfo['arbitraryfunctions']:
                try:
#                    print ('Graph plot arbitrary func', fun, type(fun))
                    f, arg, opt = fun[0], fun[1], fun[2]
                    fsplit = f.split('.')
#                    print ('   ', ax, fsplit, len(fsplit))
                    obj = ax
                    for subf in fsplit:
                        if hasattr(obj, '__call__'):
                            obj = getattr(obj(), subf)
                        else:
                            obj = getattr(obj, subf)
#                    print ('   ', obj, type(obj))
                    # handle ticks locators and formatters, take objects as arguments
                    if fsplit[-1] in ['set_major_locator', 'set_minor_locator', 'set_major_formatter', 'set_minor_formatter']:
                        if len(arg) > 0 and isinstance(arg[0], str):
                            try:
                                asplit = resplit('[()]', arg[0])
                                args = [stringToVariable(a) for a in asplit[1:] if a is not '']
                                import matplotlib.ticker as tck
                                arg = [getattr(tck, asplit[0])(*args)]
                            except Exception as e:
                                pass # just continue with unmodified arg
#                    print('      ',arg)
                    obj(*arg, **opt) # res = obj(**opt) if len(arg) == 0 else obj(*arg, **opt)
#                    print(res)
                except Exception as e:
                    print('Exception in function Graph.plot arbitrary functions')
                    print('Exception', type(e), e)
                    pass
        
        # font sizes
        if gs is None:
            listLabels = [ax.title, ax.xaxis.label, ax.yaxis.label]
            if 'ax.get_xticklabels()' not in fontsizeset:
                listLabels += ax.get_xticklabels()
            if 'ax.get_yticklabels()' not in fontsizeset:
                listLabels += ax.get_yticklabels()
            if axTwinX is not None: # maybe an automatic detection of all the axis would be more robust than trying to list all possible existing axis?
                listLabels += [axTwinX.yaxis.label] + axTwinX.get_yticklabels()
            if axTwinY is not None:
                listLabels += [axTwinY.xaxis.label] + axTwinY.get_xticklabels()
            if axTwinXY is not None:
                listLabels += [axTwinXY.xaxis.label, axTwinXY.yaxis.label] + axTwinXY.get_xticklabels() + axTwinXY.get_yticklabels()
            for item in (listLabels):
                if item not in fontsizeset:
                    item.set_fontsize(fontsize)
        
        # before saving image: restore self to its initial state
        self.update(restore)
                
        # DPI
        saveDPI = 300 if 'dpi' not in self.graphInfo else self.graphInfo['dpi']
       
        # save the graph if an export format was provided
        exportFormat = ''
        if imgFormat in ['.txt', '.xml']:
            exportFormat = imgFormat
            imgFormat = ''
        if ifSave:
            if len(imgFormat) == 0:
                imgFormat = self.config('save_imgformat', '.png')
            if not isinstance(imgFormat, list):
                imgFormat = [imgFormat]
            for imgForma_ in imgFormat:
                imgFormatTarget = ''
                if imgForma_ == '.emf': # special: we save svg and convert into emf using inkscape
                    imgFormatTarget = '.emf'
                    imgForma_ = '.svg'
                
                filename_ = filename + imgForma_
                if self.getAttribute('saveSilent') != True:
                    print('Graph saved as ' + filename_.replace('/', '\\'))
                plt.savefig(filename_, transparent=True, dpi=saveDPI)
                self.filename = filename_
                
                if imgFormatTarget == '.emf':
                    GraphIO.convertSVGtoEMF(self, filename_, imgFormat, imgFormatTarget)
        
        if ifExport:
            self.export(filesave=(filename+exportFormat))
        return [fig, axes]

            
    def convertSVGtoEMF(self, filename, imgFormat, imgFormatTarget):
        success = False
        inkscapepath = self.config('inkscape_path', [])
        if isinstance(inkscapepath, str):
            inkscapepath = [inkscapepath]
        inkscapepath += ['inkscape']
        import subprocess
        for p in inkscapepath:
            if not os.path.exists(p):
                continue # cannot find inkscape executable
            try:
                fileemf = filename[:-len(imgFormat)]+imgFormatTarget
                command = '"'+p+'" --without-gui --export-emf="'+fileemf+'" "'+filename+'"'
                out = subprocess.call(command)
                if out == 0:
                    print('Graph saved as ' + fileemf.replace('/', '\\'))
                    success = True
                    break
                else:
                    print('Graph save as .emf: likely error (return value '+out+')')
            except Exception as e:
                print('Exception during save in .emf format:', e)
                pass
        if not success:
            print('Could not save image in .emf format. Please check the following:')
            print(' - A version of inkscape is available,')
            print(' - file config.txt in grapa directory,')
            print(' - in file config.txt a line exists, similar as that, and indicate a valid inkscape executable: inkscape_path	["C:\Program Files\Inkscape\inkscape.exe"]')




# -*- coding: utf-8 -*-

"""
Created on Fri Jul 15 15:46:13 2016

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import numpy as np
import warnings
import inspect

from grapa.mathModule import is_number


class Curve:
    
    CURVE = 'Curve'

    NMTOEV = 1239.5
    LINESTYLEHIDE = ['none', 'None', None]

    def __init__ (self, Curve, attributes, silent=True):
        self.silent = silent
        self.data = np.array([])
        self.attributes = {}
        # add data
        if isinstance(Curve, (np.ndarray, np.generic)):
            self.data = Curve
        else:
            self.data = np.array(Curve)
        if self.data.shape[0] < 2 :
            self.data = self.data.reshape((2, len(self.data)))
            print ('WARNING class Curve: data Curve need to contain at least 2 columns')
            print ('   Shape of Curve:', self.data.shape)
        # clean nan pairs at end of data
        iMax = self.data.shape[1]
        for i in range(self.data.shape[1]-1, 0, -1): # loop down to 1 only, not 0
            if np.isnan(self.data[:,i]).all():
                iMax -= 1
            else:
                break
        if iMax < self.data.shape[1]:
            self.data = self.data[:,:iMax]
        # add attributes
        if isinstance(attributes, dict):
            self.update(attributes)
        else:
            print ('ERROR class Curve: attributes need to be of class dict.')

    
    # methods related to GUI
    @classmethod
    def classNameGUI(cls): # can be overridden, see CurveArrhenius
        return cls.CURVE
    
    def funcListGUI(self, **kwargs):
        """
        Fills in the Curve actions specific to the Curve type. Syntax:
        [func,
         'Button text',
         ['label 1', 'label 2', ...],
         ['value 1', 'value 2', ...],
         {'hiddenvar1': 'value1', ...}, (optional)
         [dictFieldAttributes, {}, ...]] (optional)
        By default returns quick modifs for offset and muloffset (if already
        set), and a help for some plot types (errorbar, scatter)
        In principle the current Graph and the position of self in the Graph
        can be accessed through kwargs['graph'] and kwargs['graph_i']
        """
        out = []
        # by default the offset and muloffset can be directly accessed if are already set
        typeplot = self.getAttribute('type')
        try:
            graph = kwargs['graph']
            c = kwargs['graph_i']
            if typeplot == 'errorbar': # assist functions for type errorbar
                out += self.funcListGUI_errorbar(graph, c)
            if typeplot == 'scatter': # assist function for type errorbar
                out += self.funcListGUI_scatter(graph, c)
            if typeplot.startswith('fill'): # assist function for type errorbar
                out += self.funcListGUI_fill(graph, c)
        except Exception as e:
            print('Exception', type(e), 'in Curve.funcListGUI')
        if self.getAttribute('offset', None) is not None or self.getAttribute('muloffset', None) is not None:
            out.append([self.updateValuesDictkeys, 'Modify screen offsets', ['offset', 'muloffset'], [self.getAttribute('offset'), self.getAttribute('muloffset')], {'keys': ['offset', 'muloffset']}]) # one line per function
        return out
    
    def funcListGUI_errorbar(self, graph, c):
        out = []
        types = []
        if c+1 < graph.length():
            types.append(graph.curve(c+1).getAttribute('type'))
        if c+2 < graph.length():
            types.append(graph.curve(c+2).getAttribute('type'))
        xyerrOR = ''
        if len(types) > 0:
            choices = ['', 'errorbar_xerr', 'errorbar_yerr']
            labels = ['next Curves type: ('+str(c+1)+')']+['('+str(i)+')' for i in range(c+2, c+1+len(types))]
            fieldprops = [{'field': 'Combobox', 'values': choices}] * len(labels)
            out.append([self.updateNextCurvesScatter, 'Save', labels, types, {'graph': graph, 'graph_i': c}, fieldprops])
            xyerrOR = 'OR '
        out.append([self.updateValuesDictkeys, 'Save', [xyerrOR+'error values x', 'y'], [self.getAttribute('xerr'), self.getAttribute('yerr')], {'keys': ['xerr', 'yerr']}])
        at = ['capsize', 'ecolor']
        out.append([self.updateValuesDictkeys, 'Save', at, [self.getAttribute(a) for a in at], {'keys': at}])
        return out
        
    def funcListGUI_scatter(self, graph, c):
        out = []
        types = []
        if c+1 < graph.length():
            types.append(graph.curve(c+1).getAttribute('type'))
        if c+2 < graph.length():
            types.append(graph.curve(c+2).getAttribute('type'))
        if len(types) > 0:
            choices = ['', 'scatter_c', 'scatter_s']
            labels = ['next Curves type: ('+str(c+1)+')']+['('+str(i)+')' for i in range(c+2, c+1+len(types))]
            fieldprops = [{'field':'Combobox', 'values':choices}] * len(labels)
            out.append([self.updateNextCurvesScatter, 'Save', labels, types, {'graph': graph, 'graph_i': c}, fieldprops])
        keys = ['cmap', 'vminmax']
        out.append([self.updateValuesDictkeys, 'Save', keys, [self.getAttribute(k) for k in keys], {'keys': keys}])
        return out
        
    def funcListGUI_fill(self, graph, c):
        out = []
        at = ['fill', 'hatch']
        out.append([self.updateValuesDictkeys, 'Save', at,
                    [self.getAttribute(a) for a in at], {'keys': at},
                    [{'field': 'Combobox', 'values': ['True', 'False']},
                     {'field': 'Combobox', 'values': ['', '.', '+', '/', r'\\'], 'width': 7}]])
        return out
        
        
    def updateNextCurvesScatter(self, *values, **kwargs):
        try:
            graph = kwargs['graph']
            c = kwargs['graph_i']
        except KeyError:
            print('KeyError in Curve.updateNextCurvesScatter: "graph" or "graph_i" not provided in kwargs')
        for i in range(len(values)):
            if c+i < graph.length():
                graph.curve(c+1+i).update({'type': values[i]})
        return True
        
    
    def alterListGUI(self):
        """
        Determines the possible curve visualisations. Syntax:
        ['GUI label', ['alter_x', 'alter_y'], 'semilogx']
        By default only Linear (ie. raw data) is provided
        """
        out = []
        out.append(['Linear', ['',''], ''])
        return out
    
    
    # Function related to Fits
    def updateFitParam(self, *param):
        f = self.getAttribute('_fitFunc', '')
        #print ('Curve update', param, 'f', f)
        if f != '' and not self.attributeEqual('_popt'):
            if hasattr(self, f):
                self.setY(getattr(self, f)(self.x(), *param))
                self.update({'_popt': np.array(param)})
                return True
            return 'ERROR Update fit parameter: No such fit function (',f,').'
        return 'ERROR Update fit parameter: Empty parameter (_fitFunc:',f,', _popt:',self.getAttribute('_popt'),').'
    


    # more classical class methods
    def getData(self) :
        return self.data
    
    def shape (self, idx=':') :
        if idx == ':' :
            return self.data.shape
        return self.data.shape[idx]
    
    def _fractionToFloat(self, frac_str):
        try:
            return float(frac_str)
        except ValueError:
            num, denom = frac_str.split('/')
            return float(num) / float(denom)

    def x_offsets(self, **kwargs):
        """
        Same as x(), including the effect of offset and muloffset on output.
        """
        x = self.x(**kwargs)
        offset = self.getAttribute('offset', None)
        if offset is not None:
            o = offset[0] if isinstance(offset, list) else 0
            if isinstance(o, str):
                o = self._fractionToFloat(o)
            x = x + o
        muloffset = self.getAttribute('muloffset', None)
        if muloffset is not None:
            o = muloffset[0] if isinstance(muloffset, list) else 1
            if isinstance(o, str):
                o = self._fractionToFloat(o)
            x = x * o
        return x
    
    def x (self, index=np.nan, alter='', xyValue=None, errorIfxyMix=False, neutral=False):
        """
        Returns the x data.
        index: range to the data.
        alter: possible alteration to the data (i.e. '', 'nmeV')
        xyValue: provide [x, y] value pair to be alter-ed. x, y can be np.array.
        errorIfxyMix: throw ValueError exception if alter value calculation requires both x and y components. Useful for transforming xlim and ylim, where x do not know y in advance.
        neutral: no use yet, introduced to keep same call arguments as self.y()
        """
        if alter != '':
            if alter == 'nmeV':
                if xyValue is not None: # invert order of xyValue, to help for graph xlim and ylim
                    xyValue = np.array(xyValue)
                    if len(xyValue.shape) > 1:
                        xyValue = [xyValue[0,::-1], xyValue[1,::-1]]
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore')
                    return Curve.NMTOEV / self.x(index, xyValue=xyValue)
            elif alter == 'nmcm-1':
                if xyValue is not None: # invert order of xyValue, to help for graph xlim and ylim
                    xyValue = np.array(xyValue)
                    if len(xyValue.shape) > 1:
                        xyValue = [xyValue[0,::-1], xyValue[1,::-1]]
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore')
                    return 1e7 / self.x(index, xyValue=xyValue)
            elif alter == 'MCAkeV':
                offset = self.getAttribute('_MCA_CtokeV_offset', default=0)
                mult   = self.getAttribute('_MCA_CtokeV_mult',   default=1)
                return mult * (self.x(index, xyValue=xyValue) + offset)
            elif alter == 'SIMSdepth':
                offset = self.getAttribute('_SIMSdepth_offset', default=0)
                mult   = self.getAttribute('_SIMSdepth_mult',   default=1)
                return mult * (self.x(index, xyValue=xyValue) + offset)
            elif alter != '':
                split = alter.split('.')
                if len(split) == 2:
                    module_name = 'grapa.datatypes.' + split[0][0].lower() + split[0][1:]
                    import importlib
                    try:
                        mod = importlib.import_module(module_name)
                        met = getattr(getattr(mod, split[0]), split[1])
                        return met(self, index=index, xyValue=xyValue)
                    except ImportError as e:
                        print('ERROR Curve.x Exception raised during module import', module_name+':')
                        print(e)
                else:
                    print('Error Curve.x: cannot identify alter keyword ('+alter+').')
            
        # alter might be used by subclasses
        val = self.data if xyValue is None else np.array(xyValue)
        if len(val.shape) > 1 :
            if np.isnan(index).any() :
                return val[0,:]
            return val[0,index]
        return val[0]
    
    def y_offsets(self, **kwargs):
        """
        Same as y(), including the effect of offset and muloffset on output.
        """
        y = self.y(**kwargs)
        offset = self.getAttribute('offset', None)
        if offset is not None:
            o = offset[1] if isinstance(offset, list) else offset
            if isinstance(o, str):
                o = self._fractionToFloat(o)
            y = y + o
        muloffset = self.getAttribute('muloffset', None)
        if muloffset is not None:
            o = muloffset[1] if isinstance(muloffset, list) else muloffset
            if isinstance(o, str):
               o = self._fractionToFloat(o)
            y = y * o
        return y
        
    def y (self, index=np.nan, alter='', xyValue=None, errorIfxyMix=False, neutral=False) :
        """
        Returns the y data.
        index: range to the data.
        alter: possible alteration to the data (i.e. '', 'log10abs', 'tauc')
        xyValue: provide [x, y] value pair to be alter-ed. x, y can be np.array.
        errorIfxyMix: throw ValueError exception if alter value calculation requires both x and y components. Useful for transforming xlim and ylim, where x do not know y in advance.
        neutral: additional parameter. If True prevent Jsc substraction for CurveJV.
        """
        # if some alteration required: make operation on data requested by itself, without operation
        if alter != '':
            if alter == 'log10abs':
                jsc = self.getAttribute ('Jsc') if self.getAttribute ('Jsc') != '' and not neutral else 0
                return np.log10(np.abs(self.y(index, xyValue=xyValue) + jsc))
            elif alter == 'abs':
                jsc = self.getAttribute ('Jsc') if self.getAttribute ('Jsc') != '' and not neutral else 0
                return np.abs(self.y(index, xyValue=xyValue) + jsc)
            elif alter == 'tauc':
                if errorIfxyMix:
                    return ValueError
                return np.power(self.y(index, xyValue=xyValue) * self.x(index, alter='nmeV', xyValue=xyValue), 2)
            elif alter == 'taucln1-eqe':
                if errorIfxyMix:
                    return ValueError
                return np.power(np.log(1-self.y(index, xyValue=xyValue)) * self.x(index, alter='nmeV', xyValue=xyValue), 2)
            elif alter == 'normalized':
                out = self.y(index, xyValue=xyValue)
                return out / np.max(out)
            elif alter != '':
                split = alter.split('.')
                if len(split) == 2:
                    import importlib
                    module_name = 'grapa.datatypes.' + split[0][0].lower() + split[0][1:]
                    try:
                        mod = importlib.import_module(module_name)
                        met = getattr(getattr(mod, split[0]), split[1])
                        return met(self, index=index, xyValue=xyValue)
                    except ImportError as e:
                        print('ERROR Curve.y Exception raised during module import', module_name+':')
                        print(e)
                else:
                    if alter != 'idle':
                        print('Error Curve.y: cannot identify alter keyword ('+alter+').')
#        elif alter == '-dCdlnf':
#            from mathModule import derivative
#            return - derivative(np.log(self.x(index, xyValue=xyValue)), self.y(index, xyValue=xyValue))

        # return (subset of) data
        val = self.data if xyValue is None else np.array(xyValue)
        if len(val.shape) > 1 :
            if np.isnan(index).any() :
                return val[1,:]
            return val[1,index]
        return val[1]

    def setX (self, x, index=np.nan) :
        """ Set new value for the x data. Index can be provided (self.data[0,index] = x). """
        if np.isnan(index) :
            self.data[0,:] = x
        else :
            self.data[0,index] = x
            
    def setY (self, y, index=np.nan):
        """ Set new value for the y data. Index can be provided (self.data[1,index] = y). """
        if len(self.shape()) > 1 :
            if np.isnan(index) :
                self.data[1,:] = y
            else :
                self.data[1,index] = y
        else :
            self.data[1] = y
    
    def appendPoints(self, xSeries, ySeries):
        xSeries, ySeries = np.array(xSeries), np.array(ySeries)
        if len(xSeries) == 0:
            print('Curve appendPoints: empty xSeries provided.')
            return False
        if len(ySeries) != len(xSeries):
            print('Curve appendPoints: cannot handle different data series lengths.')
            return False
        self.data = np.append(self.data, np.array([xSeries, ySeries]), axis=1)
        return True
        
    def getAttribute (self, key, default='') :
        """ Getter of attribute """
        if key.lower() in self.attributes :
            return self.attributes[key.lower()]
        return default
    
    def getAttributes(self, keysList=None):
        """ Returns all attributes, or a subset with keys given in keyList """
        if isinstance(keysList, list):
            out = {}
            for key in keysList:
                out.update({key: self.getAttribute(key)})
            return out
        return self.attributes
        
    def attributeEqual(self, key, value=''): # value='' must be same as the default parameter of method getAttribute
        """
        Check if attribute 'key' has a certain value
        The value can be left empty: check if getAttribute(key) returns the
        default value.
        """
        val = self.getAttribute(key)
        if isinstance(val, type(value)) and val == value:
            return True
        return False
        
    def update(self, attributes):
        for key in attributes :
            k = key.lower()
            if not isinstance(attributes[key], str) or attributes[key] != '':
                self.attributes.update ({k.strip(' =:\t\n').replace('ï»¿',''): attributes[key]})
            elif k in self.attributes :
                del self.attributes[k]

    def updateValuesDictkeys(self, *args, **kwargs):
        """
        Performs update({key1: value1, key2: value2, ...}) with
        arguments value1, value2, ... , (*args) and
        kwargument key=['key1', 'key2', ...]
        """
        if 'keys' not in kwargs:
            print('Error Curve updateValuesDictkeys: "keys" key must be provided, must be a list of keys corresponding to the values provided in *args.')
            return False
        if len(kwargs['keys']) != len(args):
            print('WARNING Curve updateValuesDictkeys: len of list "keys" argument must match the number of provided values (',len(args),' args provided, ',len(kwargs['keys']),' keys).')
        lenmax = min(len(kwargs['keys']), len(args))
        for i in range(lenmax):
            self.update({kwargs['keys'][i]: args[i]})
        return True
        
        
    def delete (self, key) :
        out = {}
        if key.lower() in self.attributes :
            out.update({key: self.getAttribute('key')})
            del self.attributes[key.lower()]
        return out
    
    def swapShowHide(self):
        self.update({'linestyle': '' if self.isHidden() else 'none'})
        return True
    def isHidden(self):
        return True if self.getAttribute('linestyle') in Curve.LINESTYLEHIDE else False
        


    def castCurveListGUI(self, onlyDifferent=True):
        curveTypes = [] if onlyDifferent else [Curve]
        curveTypes += Curve.__subclasses__()
        out = []
        for c in curveTypes:
            if not onlyDifferent or not isinstance(self, c):
                name = c.classNameGUI()
                out.append([name, c.__name__, c])
        key = [x[0] for x in out]
        for i in range(len(key)-2, -1, -1):
            if key[i] in key[:i]:
                del key[i]
                del out[i]
        out = [x for (y,x) in sorted(zip(key, out), key=lambda pair: pair[0])] # sort list by name
#        out.append()
        return out

    def castCurve(self, newTypeGUI):
        if newTypeGUI == 'Curve':
            newTypeGUI = 'Curve'
        if type(self).__name__ == newTypeGUI:
            return self
        curveTypes = self.castCurveListGUI()
        for typ_ in curveTypes:
            if newTypeGUI == 'Curve':
                return Curve(self.data, attributes=self.getAttributes(), silent=True)
                break
            if newTypeGUI == typ_[0] or newTypeGUI == typ_[1] or newTypeGUI.replace(' ','') == typ_[1]:
                if isinstance(self, typ_[2]):
                    print ('Curve.castCurve: already that type!')
                else:
                    return typ_[2](self.data, attributes=self.getAttributes(), silent=True)
                break
        print('Curve.castCurve: Cannot understand new type:', newTypeGUI, '(possible types:', [key for key in curveTypes],')')
        return False
        
        
    def selectData(self, xlim=None, ylim=None, alter='', offsets=False, data=None):
        """
        Returns slices of the data self.x(), self.y() which satisfy
        xlim[0] <= x <= xlim[1], and ylim[0] <= y <= ylim[1]
        alter: '', or ['', 'abs']. Affect the test AND the output.
        offset: if True, calls x and y are affected by offset and muloffset
        data [xSeries, ySeries] can be provided. Helps the code of calling
            methods being cleaner. If provided, alter and offsets are ignored.
        """
        if isinstance(alter, str):
            alter = ['', alter]
        # retrieves entry data
        if data is None:
            if offsets:
                x, y = self.x_offsets(alter=alter[0]), self.y_offsets(alter=alter[1])
            else:
                x, y = self.x(alter=alter[0]), self.y(alter=alter[1])
        else:
            x, y = np.array(data[0]), np.array(data[1])
        # identify boundaries
        if xlim is None:
            xlim = [min(x), max(x)]
        if ylim is None:
            ylim = [min(y), max(y)]
        # construct a data mask
        mask = np.ones(len(x), dtype=bool)
        for i in range(len(mask)):
            if x[i] < xlim[0] or x[i] > xlim[1]:
                mask[i] = False
            if y[i] < ylim[0] or y[i] > ylim[1]:
                mask[i] = False
        return x[mask], y[mask]
        
    
    def getPointClosestToXY(self, x, y, alter='', offsets=False):
        """
        Return the data point closest to the x,y values.
        Priority on x, compares y only if equal distance on x
        """
        if isinstance(alter, str):
            alter = ['', alter]
        # select most suitable point based on x
        datax = self.x_offsets(alter=alter[0])
        absm = np.abs(datax - x)
        idx = np.where(absm == np.min(absm))
        if len(idx) == 0:
            idx = np.argmin(absm)
        elif len(idx) == 1:
            idx = idx[0]
        else:
            # len(idx) > 1: select most suitable point based on y
            datay = self.y_offsets(index=idx, alter=alter[1])
            absM = np.abs(datay - y)
            idX = np.where(absM == np.min(absM))
            if len(idX) == 0:
                idx = idx[0]
            elif len(idX) == 1:
                idx = idx[idX[0]]
            else: # equally close in x and y -> returns first datapoint found
                idx = idx[idX[0]]
        idxOut = idx if len(idx) <= 1 else idx[0]
        if offsets:
            return self.x_offsets(index=idx)[0], self.y_offsets(index=idx)[0], idxOut # no alter, but offset for the return value
        return self.x(index=idx)[0], self.y(index=idx)[0], idxOut # no alter, no offsets for the return value
        
    def getDataCustomPickerXY(self, idx, alter='', strDescription=False):
        """
        Given an index in the Curve data, returns a modified data which has
        some sense to the user. Is overloaded in some child Curve classes, see
        for example CurveCf
        If strDescription is True, then returns a string which describes what
        this method is doing.
        """
        if strDescription:
            return '(x,y) data with offset and transform'
        if isinstance(alter, str):
            alter = ['', alter]
        attr = {}
        return self.x_offsets(index=idx, alter=alter[0]), self.y_offsets(index=idx, alter=alter[1]), attr
 
    
    def printHelpFunc(self, func, leadingstrings=None):
        """ prints the docstring of a function """
        if leadingstrings is None:
            leadingstrings = ['- ', '  ']
        a, idx = 0, None
        for line in func.__doc__.split('\n'):
            if len(line) == 0:
                continue
            if idx is None:
                idx = len(line) - len(line.lstrip(' '))
            if len(line) == idx:
                continue
            print(leadingstrings[a] + line[idx:])
            a = 1
        
        
        
    # some arithmetics
    def __add__(self, other, sub=False, interpolate=-1, offsets=False, operator='add'):
        """
        Addition operation, element-wise or interpolated along x data
        interpolate: interpolate, or not the data on the x axis
            0: no interpolation, only consider y() values
            1: interpolation, x are the points contained in self.x() or in other.x()
            2: interpolation, x are the points of self.x(), restricted to min&max of other.x()
            -1: 0 if same y values (gain time), otherwise 1
        With sub=True, performs substraction
        offsets: if True, adds Curves after computing offsets and muloffsets.
            If not adds on the raw data.
        operator: 'add', 'sub', 'mul', 'div'. sub=True is a shortcut for operatore='sub'. Overriden by sub=True argument.
        """
        def op(x, y, operator):
            if operator == 'sub':
                return x - y
            if operator == 'mul':
                return x * y
            if operator == 'div':
                return x / y
            if operator == 'pow':
                return x ** y
            if operator != 'add':
                print('WARNING Curve.__add__: unexpected operator argument(' + operator + ').')
            return x + y
        
        if sub == True:
            operator = 'sub'
        selfx = self.x_offsets if offsets else self.x
        selfy = self.y_offsets if offsets else self.y
        if not isinstance(other, Curve): # add someting/number to a Curve
            out = Curve([selfx(), op(selfy(), other, operator)], self.getAttributes())
            if offsets: # remove offset information if use it during calculation
                out.update({'offset': '', 'muloffset': ''})
            out = out.castCurve(self.classNameGUI())
            return out # cast type
        # curve1 is a Curve
        otherx = other.x_offsets if offsets else other.x
        othery = other.y_offsets if offsets else other.y
        # default mode -1: check if can gain time (avoid interpolating)
        r = range(0, min(len(selfy()), len(othery())))
        if interpolate == -1:
            interpolate = 0 if np.array_equal(selfx(index=r), otherx(index=r)) else 1
        # avoiding interpolation
        if not interpolate:
            le = min(len(selfy()), len(othery()))
            r = range(0, le)
            if le < len(selfy()):
                print('WARNING Curve __add__: Curves not same lengths, clipped result to shortest (',len(selfy()),',',len(othery()),')')
            if not np.array_equal(selfx(index=r), otherx(index=r)):
                print('WARNING Curve __add__ ('+operator+'): Curves not same x axis values. Consider interpolation (interpolate=1).')
            out = Curve([selfx(index=r), op(selfy(index=r), othery(index=r), operator)], other.getAttributes())
            out.update(self.getAttributes())
            if offsets: # remove offset information if use it during calculation
                out.update({'offset': '', 'muloffset': ''})
            out = out.castCurve(self.classNameGUI())
            return out
        else: # not elementwise : interpolate
            from scipy.interpolate import interp1d
            # construct new x -> all x which are in the range of the other curve
            datax = list(selfx())
            if interpolate == 1: # x from both self and other
                xmin = max(min(selfx()), min(otherx()))
                xmax = min(max(selfx()), max(otherx()))
                datax += [x for x in otherx() if x not in datax] # no duplicates
            else: # interpolate 2: copy x from self, restrict to min&max of other
                xmin, xmax = min(otherx()), max(otherx())
            datax = [x for x in datax if x <= xmax and x >= xmin]
            reverse = (selfx(index=0) > selfx(index=1)) if len(selfx()) > 1 else False
            datax.sort(reverse=reverse)
            f0 = interp1d( selfx(),  selfy(), kind=1)
            f1 = interp1d(otherx(), othery(), kind=1)
            datay = [op(f0(x), f1(x), operator) for x in datax]
            out = Curve([datax, datay], other.getAttributes())
            out.update(self.getAttributes())
            if offsets: # remove offset information if use it during calculation
                out.update({'offset': '', 'muloffset': ''})
            out = out.castCurve(self.classNameGUI())
            return out
            
    def __radd__(self, other, **kwargs):
        return self.__add__(other, **kwargs)
    def __sub__(self, other, **kwargs):
        """ substract operation, element-wise or interpolated """
        kwargs.update({'sub': True})
        return self.__add__(other, **kwargs)
    def __rsub__(self, other, **kwargs):
        """ reversed substract operation, element-wise or interpolated """
        kwargs.update({'sub': False, 'operator': 'add'})
        return Curve.__add__(self.__neg__(), other, **kwargs)
    def __mul__(self, other, **kwargs):
        """ multiplication operation, element-wise or interpolated """
        kwargs.update({'operator': 'mul'})
        return self.__add__(other, **kwargs)
    def __rmul__(self, other, **kwargs):
        kwargs.update({'operator': 'mul'})
        return self.__add__(other, **kwargs)
    def __div__(self, other, **kwargs):
        """ division operation, element-wise or interpolated """
        kwargs.update({'operator': 'div'})
        return self.__add__(other, **kwargs)
    def __rdiv__(self, other, **kwargs):
        """ division operation, element-wise or interpolated """
        kwargs.update({'operator': 'mul'})
        return Curve.__add__(self.__invertArithmetic__(), other, **kwargs)
    def __truediv__(self, other, **kwargs):
        """ division operation, element-wise or interpolated """
        kwargs.update({'operator': 'div'})
        return self.__add__(other, **kwargs)
    def __rtruediv__(self, other, **kwargs):
        """ division operation, element-wise or interpolated """
        kwargs.update({'operator': 'mul'})
        return Curve.__add__(self.__invertArithmetic__(), other, **kwargs)
    def __pow__(self, other, **kwargs):
        """ power operation, element-wise or interpolated """
        kwargs.update({'operator': 'pow'})
        return self.__add__(other, **kwargs)

    def __neg__(self, **kwargs):
        out = Curve([self.x(), -self.y()], self.getAttributes())
        out = out.castCurve(self.classNameGUI())
        return out
    def __invertArithmetic__(self, **kwargs):
        out = Curve([self.x(), 1/self.y()], self.getAttributes())
        out = out.castCurve(self.classNameGUI())
        return out
        
        


    def plot(self, ax, graph=None, graph_i=None, type_plot='', ignoreNext=0, boxplot=None, violinplot=None, violinplotkwargs={}):
        """
        plot a Curve on some axis
        graph, graph_i: a graph instance, such that self==graph.curve(graph_i)
            required to properly plot scatter with scatter_c, etc.
        alter: '', or ['nmeV', 'abs']
        type_plot: 'semilogy'
        ignoreNext: int, counter to decide whether the next curves shall not be
            plotted (multi-Curve plotting such as scatter)
        boxplot:
        violinplot:
        violinplotkwargs:
        """
        from grapa.graph import Graph
        handle = None
        # check default arguments
        if boxplot is None:
            boxplot    = {'y':[], 'positions':[], 'labels':[], 'color':[], 'i':0}
        if violinplot is None:
            violinplot = {'y':[], 'positions':[], 'labels':[], 'color':[]}
        if graph is None:
            graph_i = None
        else:
            if graph.curve(graph_i) != self:
                graph_i = None
            if graph_i is None:
                for c in range(graph.length()):
                    if graph.curve(c) == self:
                        graph_i = c
                        break
            if graph_i is None:
                graph = None # self was not found in graph
                print('Warning Curve.plot: Curve not found in provided Graph')
                
                
        # retrieve basic information
        alter = graph._getAlter() if graph is not None else ['', '']
        attr = self.getAttributes()
        linespec = self.getAttribute('linespec')
        # construct dict of keywords based on curves attributes, in a very restrictive way
        # some attributes are commands for plotting, some are just related to the sample, and no obvious way to discriminate between the 2
        fmt = {}
        for key in attr:
            if not isinstance(key, str):
                print(type(key), key, attr[key])
            if (not isinstance(attr[key], str) or attr[key] != '') and key in Graph.dataInfoKeysGraph and key not in ['plot', 'linespec', 'type', 'ax_twinx', 'ax_twiny', 'offset', 'muloffset', 'labelhide', 'colorbar', 'xerr', 'yerr']:
                fmt[key] = attr[key]
        # do not plot curve if was asked not to display it.
        if 'linestyle' in fmt and fmt['linestyle'] in Curve.LINESTYLEHIDE:
            return None, ignoreNext
        # some renaming of kewords, etc
        if 'legend' in fmt:
            fmt['label'] = fmt['legend']
            del fmt['legend']
        if 'cmap' in fmt and not isinstance(fmt['cmap'], str):
             # convert Colorscale into matplotlib cmap
            from grapa.colorscale import Colorscale
            fmt['cmap'] = Colorscale(fmt['cmap']).cmap()
        if 'vminmax' in fmt:
            if isinstance(fmt['vminmax'], list) and len(fmt['vminmax']) > 1:
                if fmt['vminmax'][0] != '' and not np.isnan(fmt['vminmax'][0]) and not np.isinf(fmt['vminmax'][0]):
                    fmt.update({'vmin': fmt['vminmax'][0]})
                if fmt['vminmax'][1] != '' and not np.isnan(fmt['vminmax'][1]) and not np.isinf(fmt['vminmax'][1]):
                    fmt.update({'vmax': fmt['vminmax'][1]})
            del fmt['vminmax']

        # start plotting
        # retrieve data after transform, including of offset and muloffset
        x = self.x_offsets(alter=alter[0])
        y = self.y_offsets(alter=alter[1])
        type_graph = self.getAttribute('type', 'plot')
        if type_plot.endswith(' norm.'):
            type_graph = type_plot[:-6]
            y = y / max(y)
        
        # add keyword arguments which are in the plot method prototypes
        try:
            sig = inspect.signature(getattr(ax, type_graph))
            for key in sig.parameters:
                if key in attr and key not in fmt:
                    fmt.update({key: attr[key]})
        except AttributeError:
            print('Curve.plot: desired plotting method not found ('+type_graph+'). Going for default.')
            pass # for xample 'errorbar_yerr' after suppression of previous Curve 'errorbar'. Will be 'plot' anyway.
        except Exception as e:
            print('Exception in Curve.plot while identifying keyword arguments:')
            print(type(e), e)
        
        if 'labelhide' in attr and attr['labelhide']:
            if 'label' in fmt:
                del fmt['label']

        # No support for the following methods (either 2D data, or complicated to implement):
        #    hlines, vlines, broken_barh, contour, contourf, polar,
        #    pcolor, pcolormesh, streamplot, tricontour, tricontourf,
        #    tripcolor
        # Partial support for:
        #    imgshow
        
        # "simple" plotting methods, with prototype similar to plot()
        if type_graph in ['fill', 'semilogx', 'semilogy', 'loglog', 'plot_date', 'stem', 'step', 'triplot']:
            handle = getattr(ax, type_graph)(x, y, linespec, **fmt)
        # plotting methods not accepting formatting string as 3rd argument
        elif type_graph in ['bar', 'barbs', 'barh', 'cohere', 'csd', 'fill_between', 'fill_betweenx', 'hexbin', 'hist2d', 'quiver', 'xcorr']:
            handle = getattr(ax, type_graph)(x, y, **fmt)
        #  plotting of single vector data
        elif type_graph in ['acorr', 'angle_spectrum', 'eventplot', 'hist', 'magnitude_spectrum', 'phase_spectrum', 'pie', 'psd', 'specgram']:
            # careful with eventplot, the Curve data are modified
            handle = getattr(ax, type_graph)(y, **fmt)
        # a more peculiar plotting
        elif type_graph in ['spy']:
            handle = getattr(ax, type_graph)([x, y], **fmt)
        elif type_graph == 'stackplot':
            # look for next Curves with type == 'stackplot', and same x
            nexty = []
            fmt['labels'] = [fmt['label']]
            del fmt['label']
            if graph is not None:
                for j in range(graph_i+1, graph.length()):
                    if graph.curve(j).getAttribute('type') == type_graph and np.array_equal(x, graph.curve(j).x_offsets(alter=alter[0])):
                        ignoreNext += 1
                        nexty.append(graph.curve(j).y_offsets(alter=alter[1]))
                        fmt['labels'].append(graph.curve(j).getAttribute('label'))
                        continue
                    else:
                        break
            handle = getattr(ax, type_graph)(x, y, *nexty, **fmt)
        elif type_graph == 'errorbar':
            # look for next Curves, maybe xerr/yerr was provided
            if 'xerr' in attr:
                fmt.update({'yerr': attr['xerr']})
            if 'yerr' in attr:
                fmt.update({'yerr': attr['yerr']})
            if graph is not None:
                for j in range(graph_i+1, min(graph_i+3, graph.length())):
                    if len(graph.curve(j).y()) == len(y):
                        typenext = graph.curve(j).getAttribute('type')
                        if typenext not in ['errorbar_xerr', 'errorbar_yerr']:
                            break
                        if typenext == 'errorbar_xerr':
                            fmt.update({'xerr': graph.curve(j).y_offsets()})
                            ignoreNext += 1
                            continue
                        if typenext == 'errorbar_yerr':
                            fmt.update({'yerr': graph.curve(j).y_offsets()})
                            ignoreNext += 1
                            continue
                    break
            handle = ax.errorbar(x, y, fmt=linespec, **fmt)
        elif type_graph == 'scatter':
            convert = {'markersize': 's', 'markeredgewidth': 'linewidths'}
            for key in convert:
                if key in fmt:
                    fmt.update({convert[key]: fmt[key]})
                    del fmt[key]
            try:
                if graph is not None:
                    for j in range(graph_i+1, min(graph_i+3, graph.length())):
                        typenext = graph.curve(j).getAttribute('type')
                        if typenext not in ['scatter_c', 'scatter_s']:
                            break
                        if 's' not in fmt and typenext == 'scatter_s':
                            fmt.update({'s': graph.curve(j).y_offsets(alter=alter[1])})
                            ignoreNext += 1
                            continue
                        elif 'c' not in fmt and (typenext == 'scatter_c' or np.array_equal(x, graph.curve(j).x_offsets(alter=alter[0]))):
                            fmt.update({'c': graph.curve(j).y_offsets(alter=alter[1])})
                            ignoreNext += 1
                            if 'color' in fmt:
                                del fmt['color'] # there cannot be both c and color keywords
                            continue
                        else:
                            break
                handle = ax.scatter(x, y, **fmt)
            except Exception as e:
                print('ERROR! Exception occured in Curve.plot function during scatter.')
                print(type(e), e)
        elif type_graph == 'boxplot':
            if len(y) > 0 and not np.isnan(y).all():
                boxplot['y'].append(y[~np.isnan(y)])
                boxplot['positions'].append(boxplot['i'])
                boxplot['labels'].append(fmt['label'] if 'label' in fmt else '')
                boxplot['color'].append (fmt['color'] if 'color' in fmt else '')
                boxplot['i'] += 1
        elif type_graph == 'violinplot':
            if len(y) > 0 and not np.isnan(y).all():
                violinplot['y'].append(y[~np.isnan(y)])
                violinplot['positions'].append(boxplot['i'])
                violinplot['labels'].append(fmt['label'] if 'label' in fmt else '')
                violinplot['color'].append (fmt['color'] if 'color' in fmt else '')
                if 'showmeans' in attr:
                    violinplotkwargs.update({'showmeans':   attr['showmeans']})
                if 'showmedians' in attr:
                    violinplotkwargs.update({'showmedians': attr['showmedians']})
                if 'showextrema' in attr:
                    violinplotkwargs.update({'showextrema': attr['showextrema']})
                boxplot['i'] += 1
        elif type_graph in ['imshow', 'contour', 'contourf']:
            from grapa.curve_image import Curve_Image
            img, ignoreNext, X, Y = Curve_Image.getImageData(self, graph, graph_i, alter, ignoreNext)
            if 'label' in fmt:
                del fmt['label']
            if type_graph in ['contour', 'contourf']:
                for key in ['corner_mask', 'colors', 'alpha', 'cmap', 'norm', 'vmin', 'vmax', 'levels', 'origin', 'extent', 'locator', 'extend', 'xunits', 'yunits', 'antialiased', 'nchunk', 'linewidths', 'linestyles', 'hatches']:
                    if key in attr and key not in fmt:
                        fmt.update({key: attr[key]})
                # TODO: remove linewidths, linestyles for contourf, hatches for contour
            args = [img]
            if X is not None and Y is not None and type_graph in ['contour', 'contourf']:
                args = [X, Y] +  args
            try:
                handle = getattr(ax, type_graph)(*args, **fmt)
            except Exception as e:
                print('Curve plot', type_graph, 'Exception')
                print(type(e), e)
        else: # default is plot (lin-lin) # also valid if no information is stored, aka returned ''
            handle = ax.plot(x, y, linespec, **fmt)
    
        handles = handle if isinstance(handle, list) else [handle]
        for key in attr:
            if key not in fmt and key not in ['label', 'plot', 'linespec', 'type', 'ax_twinx', 'ax_twiny', 'offset', 'muloffset', 'labelhide', 'colorbar']:
                for h in handles:
                    if hasattr(h, 'set_'+key):
                        try:
                            getattr(h, 'set_'+key)(attr[key])
                        except Exception as e:
                            print('GraphIO Exception during plot kwargs adjustment for key', key, ':', type(e))
                            print(e)
        
        return handle, ignoreNext
        
        



from setuptools import setup


with open('readme.rst', 'r') as f:
    long_description = f.read()


setup(
    name='grapa',
    version='0.5.0.0',
    description='Grapa - graphing and photovoltaics analysis',
    author='Romain Carron',
    author_email='carron.romain@gmail.com',
    install_requires=[
        'numpy',
        'matplotlib',
        'scipy',
    ],
    packages=[
        'grapa',
        'grapa.datatypes',
        'grapa.scripts',
        'grapa.gui',
    ],
    license='MIT',
    url='https://github.com/carronr/grapa/',
)



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 27 10:18:06 2017

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import numpy as np
from grapa.graph import Graph
from grapa.curve import Curve

class GraphMydata(Graph):
    # will be used by grapa code
    FILEIO_GRAPHTYPE = 'MyCurve'
    # 3-element list, default x and y axis labels: [text, symbol, unit]
    AXISLABELS = [['theta', '\theta', 'unit'], ['value', '', 'unit']]
    
    @classmethod
    def isFileReadable(cls, fileName, fileExt, line1='', line2='', line3='', **kwargs):
        """
        This function must return True if the file is recognized as of your data format.
        Please be selective so as not to open other file formats by mistake.
        Are provided the file name, extension, and the 3 first lines of the file
        """
        print('File open GraphMydata', (fileExt == '.txt'), 'line1', line1)
        if fileExt == '.txt' and line1 == 'This is my own data format':
            return True
        return False

    def readDataFromFile(self, attributes, **kwargs):
        """
        This function parses your files
        The file can be opened using open(self.filename, 'r')
        """
        print('File open GraphMydata')
        # interpret headers - here only attribute 'label'
        attrs = {}
        f = open(self.filename, 'r')
        f.readline()
        attrs['label'] = f.readline().strip()
        f.close()
        # read data
        data = np.array([[np.nan], [np.nan]])
        kw = {'skip_header': 2, 'usecols': [0, 1]}
        try:
            data = np.transpose(np.genfromtxt(self.filename, **kw))
        except Exception as e:
            if not self.silent:
                print('WARNING cannot read file', self.filename)
                print(type(e), e)
        # create Curve object(s)
        try:
            from grapa.datatypes.curveMycurve import CurveMycurve
            self.append(CurveMycurve(data, attributes))
        except ImportError:
            self.append(Curve(data, attributes))
        self.curve(-1).update(attrs) # file content may override default attributes
        
        # cosmetics
        # some default settings
        self.update({'ylim': [-1.2, 1.2],
                     'xlabel': self.formatAxisLabel(GraphMydata.AXISLABELS[0]),
                     'ylabel': self.formatAxisLabel(GraphMydata.AXISLABELS[1])})



# -*- coding: utf-8 -*-
"""
Created on Sun Oct 29 15:03:50 2017

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

from grapa.graph import Graph
from grapa.curve import Curve

class Curve_Subplot(Curve):
    """
    The purpose is this class is to provid GUI support to handle subplots
    """

    CURVE = 'subplot'

    def __init__(self, *args, **kwargs):
        Curve.__init__(self, *args, **kwargs)
        # define default values for important parameters
        self.update ({'Curve': Curve_Subplot.CURVE})
        
        def checkattr(key, default, typ_):
            val = self.getAttribute(key)
            if val is '':
                self.update({key: default})
            elif not isinstance(val, typ_):
                self.update({key: typ_(val)})
        
        checkattr('subplotfile', ' ', str)
        checkattr('subplotrowspan', 1, int)
        checkattr('subplotcolspan', 1, int)
        checkattr('subplotupdate', {}, dict)
        if self.getAttribute('subplotfile') is ' ':
            if self.getAttribute('insetfile') not in ['', ' ']:
                self.update({'subplotfile': self.getAttribute('insetfile')})
                self.update({'insetfile': ''})

    # GUI RELATED FUNCTIONS
    def funcListGUI(self, **kwargs):
        out = Curve.funcListGUI(self, **kwargs)
        graph = None
        if 'graph' in kwargs:
            graph = kwargs['graph']
            del kwargs['graph']
        out.append([self.updateValuesDictkeys, 'Set', ['file subplot'], [self.getAttribute('subplotfile')],    {'keys': ['subplotfile']}])
        dictspan = {'field':'Combobox', 'values':['', '1','2','3','4','1 number'], 'bind':'beforespace'}
        out.append([self.updateValuesDictkeys, 'Set', ['colspan', 'rowspan'],
                    [self.getAttribute('subplotcolspan'), self.getAttribute('subplotrowspan')],
                    {'keys': ['subplotcolspan', 'subplotrowspan']},
                    [dictspan, dictspan]])
        out.append([self.updateValuesDictkeys, 'Set', ['update subplot'], [self.getAttribute('subplotupdate')],  {'keys': ['subplotupdate']}])
        if graph is not None and isinstance(graph, Graph):
            out.append([None, 'Graph options', [], []])
            ncols = int(graph.getAttribute('subplotsncols', 2))
            out.append([graph.updateValuesDictkeys, 'Set',
                        ['n cols', 'transpose', 'show id? or ('+u"\u0394x,\u0394y)"],
                        [ncols, str(graph.getAttribute('subplotstranspose')), str(graph.getAttribute('subplotsid'))],
                        {'keys': ['subplotsncols', 'subplotstranspose', 'subplotsid']},
                        [{'field':'Combobox', 'values':['','1','2','3','4','1 number']},
                         {'field':'Combobox', 'values':['', '0 False', '1 True'], 'bind':'beforespace'},{}]])
            out.append([graph.updateValuesDictkeys, 'Set', ['width ratios (i.e. [1,2,1])', 'height_ratios'], [graph.getAttribute('subplotswidth_ratios'), graph.getAttribute('subplotsheight_ratios')], {'keys': ['subplotswidth_ratios', 'subplotsheight_ratios']}])
            out.append([graph.updateValuesDictkeys, 'Set', ['subplots_adjust [left, bottom, right, top, wspace, hspace]'], [graph.getAttribute('subplots_adjust')], {'keys': ['subplots_adjust']}])
        out.append([self.printHelp, 'Help!', [], []]) # one line per function
        return out
    

        
    def printHelp(self):
        print('*** *** ***')
        print('Class Curve_Subplot facilitates the creation and customization of subplots inside a Graph.')
        print('The main graph will be subdivised in an array of subplots.')
        print('Important parameters:')
        print('- subplotfile: saved Graph file which must be shown in subplot.')
        print('  The next Curves will also be shown in the created axis, until the next Curve with same Curve_Subplot type.')
        print('- subplotcolspan: on how many column the subplot will be plotted.')
        print('- subplotrowspan: on how many rows the subplot will be plotted.')
        print('- subplotupdate: a dict which will be applied to the subplot Graph. Provides basic support for run-time customization of the inset graph.')
        print("  Examples: {'fontsize':8}, or {'xlabel': 'An updated label'}")
        print()
        return True

    


# -*- coding: utf-8 -*-
"""
Created on Mon Dec 12 20:08:04 2016

@author: car
Copyright (c) 2018, Empa, Romain Carron
"""

#print ('matplotlib 4')
import matplotlib.pyplot as plt
import os
import numpy as np
import sys

folder = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', '..'))
if folder not in sys.path:
    sys.path.append(folder)
from grapa.graph import Graph



def isSimilar(a, b, tolRel=1e-5):
    if np.max(np.abs((np.array(a) - np.array(b)) / np.array(b))) < tolRel:
        return True
    print('Relative difference', (np.array(a) - np.array(b))/np.array(b))
    return False
def test(result, expected, tolRel, description):
    if isSimilar(result, expected, tolRel=tolRel):
        print('    OK', description)
        return True
    print('    ERROR', description, result, 'expected', expected)
    return False

        
def testJscVoc(graphs=True):
    ret = True
    from grapa.datatypes.curveJscVoc import GraphJscVoc
    graph = Graph('JscVoc/JscVoc_SAMPLE_a3_Values.txt', silent=True)
    out = GraphJscVoc.CurvesJscVocSplitTemperature(graph, threshold=3, curve=graph.curve(0))
    for c in out:
        c.update({'linespec': '.'})
    graph.append(out)
    
    ret *= test(graph.curve(-1).fit_nJ0(), [4.55823, 0.002558908], tolRel=1e-4, description='JscVoc fit single T (1)')
    
    out = GraphJscVoc.CurveJscVoc_fitNJ0(graph, curve=graph.curve(-1), silent=True, Jsclim=[0.5,np.inf])
    ret *= test(out[0].getAttribute('_popt'), [1.745988, 4.0931023e-8], tolRel=1e-4, description='JscVoc fit single T (2)')
    
    out = GraphJscVoc.CurveJscVoc_fitNJ0(graph, Jsclim=[0.5,np.inf], curve=graph.curve(0), graphsnJ0=False, silent=True)
    ret *= test(out[-1].getAttribute('_popt'), [1.745988, 4.0931023e-8], tolRel=1e-4, description='JscVoc fit multiple T')
    
    out = GraphJscVoc.CurveJscVoc_fitNJ0(graph, Jsclim=[1,np.inf], curve=graph.curve(0), graphsnJ0=True, silent=True)
    graph.append(out)
    if graphs:
        graph.plot(ifSave=False, ifExport=False)
    for c in graph.iterCurves():
        c.swapShowHide()
    graph.curve(-2).swapShowHide()
    graph.curve(-3).swapShowHide()
    graph.update({'typeplot': ''})
    graph.update({'alter': ['CurveArrhenius.x_1000overK', 'CurveArrhenius.y_Arrhenius']})
    #graph.update({'xlim': [np.inf, 350], 'ylim': [-30,10]})
    out = graph.curve(-1).CurveArrhenius_fit([-np.inf, np.inf], silent=True)
    graph.append(out)
    ret *= test(out.getAttribute('_popt'), [1.4993444970660568, 656784.49095749843], tolRel=1e-4, description='JscVoc J0 Arrhenius')
    if graphs:
        graph.plot(ifSave=False, ifExport=False)
    return ret
    

def testEQEAdd(graphs=True):
    ret = True
    graph = Graph('./EQE/SAMPLE_A_d1_1.sr', silent=True)
    target = 32.9679429444
    ret *= test(graph.curve(0).currentCalc(silent=True), target, tolRel=1e-4, description='EQE current')
    if graphs:
        graph.plot(ifSave=False, ifExport=False)
    return ret

    




def testAll():
    graphs = False
    
    # run various tests
    out = testJscVoc(graphs=graphs)
    print('test JscVoc: ', 'passed' if out else 'FAILED')

    out = testEQEAdd(graphs=graphs)
    print('test EQE: ', 'passed' if out else 'FAILED')

    
    # all scripts can be check independently
    # -> JV, CV, Cf, EQE ok
    # TODO: Spectrum, TRPL, SIMS, MCA
    # ARRHENIUS done in JscVoc
    
    plt.show()
    
    
if __name__ == "__main__":
    testAll()
